'''
 Copyright 2022 Yingte Xu
 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at

 http://www.apache.org/licenses/LICENSE-2.0

 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.
'''

# ------------------------------------------------------------
# opt_term.py
#
# defining the operator terms, and their corresponding properties
# ------------------------------------------------------------
from __future__ import annotations
from typing import Any, List, Tuple

from . import opt_kernel

from nqpv import dts

import numpy as np

fac = dts.TermFact()

type_operator = fac.axiom("operator", fac.sort_term(0))

class OperatorTerm(dts.Term):
    def __init__(self, m : np.ndarray):
        if not isinstance(m, np.ndarray):
            raise ValueError()

        super().__init__(type_operator, None)
        self._m : np.ndarray = m
        self._qnum : int | None = None
        self._unitary : bool | None = None
        self._hermitian_predicate : bool | None = None
        self._measurement : bool | None = None
    
    @property
    def m(self) -> np.ndarray:
        return self._m

    @property
    def qnum(self) -> int:
        if self._qnum is None:
            self._qnum = opt_kernel.get_opt_qnum(self._m)
        return self._qnum

    def ensure_unitary(self) -> None:
        self._unitary = True
    
    def ensure_hermitian_predicate(self) -> None:
        self._hermitian_predicate = True
    
    def ensure_measurement(self) -> None:
        self._measurement = True

    @property
    def unitary(self) -> bool:
        if self._unitary is None:
            self._unitary = opt_kernel.check_unity(self._m)
        return self._unitary

    @property
    def hermitian_predicate(self) -> bool:
        if self._hermitian_predicate is None:
            self._hermitian_predicate = opt_kernel.check_hermitian_predicate(self._m)
        return self._hermitian_predicate
    
    @property
    def measurement(self) -> bool:
        if self._measurement is None:
            self._measurement = opt_kernel.check_measure(self._m)
        return self._measurement
    
    def __eq__(self, other) -> bool:
        if isinstance(other, OperatorTerm):
            return opt_kernel.np_eps_equal(self._m, other._m)
        elif isinstance(other, dts.Var):
            raise NotImplemented
        else:
            return False
    
    def __str__(self) -> str:
        # output according to its property
        if self.hermitian_predicate or self.unitary:
            return str(self._m.reshape((2**self.qnum, 2**self.qnum)))
        elif self.measurement:
            return str(self._m.reshape((2, 2**self.qnum, 2**self.qnum)))
        else:
            return str(self._m)

    def dagger(self) -> OperatorTerm:
        '''
        return the dagger operator
        '''
        r = OperatorTerm(opt_kernel.dagger(self._m))
        if self.unitary:
            r.ensure_unitary()
        if self.hermitian_predicate:
            r.ensure_hermitian_predicate()
        return r

    def opt_mea0(self) -> OperatorTerm:
        '''
        return the operator for measurement result 0
        '''
        if not self.measurement:
            raise ValueError()
        return OperatorTerm(self._m[0])
    
    def opt_mea1(self) -> OperatorTerm:
        '''
        return the operator for measurement result 1
        '''
        if not self.measurement:
            raise ValueError()
        return OperatorTerm(self._m[1])

    def __add__(self, other : OperatorTerm) -> OperatorTerm:
        '''
        return the addition result
        (cannot automatically ensure any property)
        '''
        if not isinstance(other, OperatorTerm):
            raise ValueError()
        r = OperatorTerm(self._m + other._m)
        '''
        # check whether it is hermitian predicate
        if self.hermitian_predicate and other.hermitian_predicate:
            r.ensure_hermitian_predicate()
        '''
        return r

def val_opt(term : dts.Term) -> OperatorTerm:
    if not isinstance(term, dts.Term):
        raise ValueError()
    if term.type != type_operator:
        raise ValueError()

    if isinstance(term, OperatorTerm):
        return term
    elif isinstance(term, dts.Var):
        val = term.val
        if not isinstance(val, OperatorTerm):
            raise Exception()
        return val
    else:
        raise Exception()



'''
 Copyright 2022 Yingte Xu
 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at

 http://www.apache.org/licenses/LICENSE-2.0

 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.
'''

# ------------------------------------------------------------
# proof_hint_term.py
#
# define the proof hint terms
# ------------------------------------------------------------

from __future__ import annotations
from typing import List, Tuple

from nqpv import dts
from nqpv.vsystem.log_system import RuntimeErrorWithLog

from .qvarls_term import QvarlsTerm, type_qvarls, val_qvarls
from .opt_pair_term import OptPairTerm, type_opt_pair, val_opt_pair
from . import qpre_term
from .qpre_term import QPreTerm, type_qpre, val_qpre
from .scope_term import ScopeTerm

from .proof_term import type_proof, ProofDefinedTerm, val_proof

fac = dts.TermFact()
type_proof_hint = fac.axiom("proof_hint", fac.sort_term(0))

class ProofHintTerm(dts.Term):
    def __init__(self, all_qvarls : QvarlsTerm, label : str):
        val_qvarls(all_qvarls)

        super().__init__(type_proof_hint, None)
        self._all_qvarls : QvarlsTerm = all_qvarls
        self._label : str = label

    @property
    def all_qvarls(self) -> QvarlsTerm:
        return self._all_qvarls
    
    @property
    def label(self) -> str:
        return self._label
    
    def prog_consistent(self, other : ProofHintTerm) -> bool:
        '''
        check whether the two proof hints are about the same program (syntactically)
        '''
        raise NotImplementedError()

    def str_content(self, prefix : str) -> str:
        raise NotImplementedError()

    def __str__(self) -> str:
        return "\n" + self.str_content("") + "\n"
    
def val_proof_hint(term : dts.Term) -> ProofHintTerm:
    if not isinstance(term, dts.Term):
        raise ValueError()
    if term.type != type_proof_hint:
        raise ValueError()
        
    if isinstance(term, ProofHintTerm):
        return term
    elif isinstance(term, dts.Var):
        val = term.val
        if not isinstance(val, ProofHintTerm):
            raise Exception()
        return val
    else:
        raise Exception()

class SkipHintTerm(ProofHintTerm):
    def __init__(self):
        super().__init__(QvarlsTerm(()), "skip hint")

    def prog_consistent(self, other: ProofHintTerm) -> bool:
        return isinstance(other, SkipHintTerm)

    def str_content(self, prefix: str) -> str:
        return prefix + "skip"

class AbortHintTerm(ProofHintTerm):
    def __init__(self):
        super().__init__(QvarlsTerm(()), "abort hint")

    def prog_consistent(self, other: ProofHintTerm) -> bool:
        return isinstance(other, AbortHintTerm)

    def str_content(self, prefix: str) -> str:
        return prefix + "abort"

class InitHintTerm(ProofHintTerm):
    def __init__(self, qvarls : dts.Term):
        if not isinstance(qvarls, dts.Term):
            raise ValueError()
        if qvarls.type != type_qvarls:
            raise RuntimeErrorWithLog("The term '" + str(qvarls) + "' is not a quantum variable list.")
        
        qvarls_val = val_qvarls(qvarls)

        super().__init__(qvarls_val, "initialization hint")
        self._qvarls : dts.Term = qvarls

    @property
    def qvarls_val(self) -> QvarlsTerm:
        return val_qvarls(self._qvarls)

    def prog_consistent(self, other: ProofHintTerm) -> bool:
        if isinstance(other, InitHintTerm):
            return self._qvarls == other._qvarls
        else:
            return False

    def str_content(self, prefix: str) -> str:
        return prefix + str(self._qvarls) + " :=0"

class UnitaryHintTerm(ProofHintTerm):
    def __init__(self, opt_pair : dts.Term):
        if not isinstance(opt_pair, dts.Term):
            raise ValueError()
        if opt_pair.type != type_opt_pair:
            raise RuntimeErrorWithLog("The term '" + str(opt_pair) + "' is not a operator variable pair.")
        opt_pair_val = val_opt_pair(opt_pair)
        if not opt_pair_val.unitary_pair:
            raise RuntimeErrorWithLog("The operator variable pair '" + str(opt_pair) + "' is not an unitary pair.")
        
        all_qvarls = opt_pair_val.qvarls_val
        super().__init__(all_qvarls, "unitary hint")
        self._opt_pair : dts.Term = opt_pair

    @property
    def opt_pair_val(self) -> OptPairTerm:
        return val_opt_pair(self._opt_pair)

    def prog_consistent(self, other: ProofHintTerm) -> bool:
        if isinstance(other, UnitaryHintTerm):
            return self._opt_pair == other._opt_pair
        else:
            return False

    def str_content(self, prefix: str) -> str:
        return prefix + str(self.opt_pair_val._qvarls) + " *= " + str(self.opt_pair_val._opt)

class IfHintTerm(ProofHintTerm):
    def __init__(self, opt_pair : dts.Term, P1 : dts.Term, P0 : dts.Term):
        if not isinstance(opt_pair, dts.Term) or not isinstance(P0, dts.Term) or not isinstance(P1, dts.Term):
            raise ValueError()
        # check the measurement
        if opt_pair.type != type_opt_pair:
            raise RuntimeErrorWithLog("The term '" + str(opt_pair) + "' is not a operator variable pair.")
        opt_pair_val = val_opt_pair(opt_pair)
        if not opt_pair_val.measurement_pair:
            raise RuntimeErrorWithLog("The operator variable pair '" + str(opt_pair) + "' is not a measurement set pair.")

        if P0.type != type_proof_hint:
            raise RuntimeErrorWithLog("The term '" + str(P0) + "' is not a proof hint.")
        if P1.type != type_proof_hint:
            raise RuntimeErrorWithLog("The term '" + str(P1) + "' is not a proof hint.")
        
        P0_val = val_proof_hint(P0)
        P1_val = val_proof_hint(P1)
        
        all_qvarls = opt_pair_val.qvarls_val
        all_qvarls = all_qvarls.join(P0_val.all_qvarls)
        all_qvarls = all_qvarls.join(P1_val.all_qvarls)
        super().__init__(all_qvarls, "if hint")
        self._opt_pair : dts.Term = opt_pair
        self._P0 : dts.Term = P0
        self._P1 : dts.Term = P1

    @property
    def opt_pair_val(self) -> OptPairTerm:
        return val_opt_pair(self._opt_pair)
    
    @property
    def P0_val(self) -> ProofHintTerm:
        return val_proof_hint(self._P0)
    
    @property
    def P1_val(self) -> ProofHintTerm:
        return val_proof_hint(self._P1)

    def prog_consistent(self, other: ProofHintTerm) -> bool:
        if isinstance(other, IfHintTerm):
            return self._opt_pair == other._opt_pair\
                and self.P1_val.prog_consistent(other.P1_val)\
                and self.P0_val.prog_consistent(other.P0_val)
        else:
            return False

    def str_content(self, prefix: str) -> str:
        r = prefix + "if " + str(self._opt_pair) + " then\n"
        r += self.P1_val.str_content(prefix + "\t") + "\n"
        r += prefix + "else\n"
        r += self.P0_val.str_content(prefix + "\t") + "\n"
        r += prefix + "end"
        return r

class WhileHintTerm(ProofHintTerm):
    def __init__(self, inv : dts.Term, opt_pair : dts.Term, P : dts.Term):
        if not isinstance(inv, dts.Term) or not isinstance(opt_pair, dts.Term) or not isinstance(P, dts.Term):
            raise ValueError()
        # check the measurement
        if opt_pair.type != type_opt_pair:
            raise RuntimeErrorWithLog("The term '" + str(opt_pair) + "' is not a operator variable pair.")
        opt_pair_val = val_opt_pair(opt_pair)
        if not opt_pair_val.measurement_pair:
            raise RuntimeErrorWithLog("The operator variable pair '" + str(opt_pair) + "' is not a measurement set pair.")
        
        # check loop invariant
        if inv.type != type_qpre:
            raise RuntimeErrorWithLog("The term '" + str(opt_pair) + "' is not a predicate, while a loop invariant is needed.")

        if P.type != type_proof_hint:
            raise RuntimeErrorWithLog("The term '" + str(P) + "' is not a proof hint.")

        P_val = val_proof_hint(P)
        
        all_qvarls = opt_pair_val.qvarls_val
        all_qvarls = all_qvarls.join(P_val.all_qvarls)
        super().__init__(all_qvarls, "while hint")
        self._inv : dts.Term = inv
        self._opt_pair : dts.Term = opt_pair
        self._P : dts.Term = P

    @property
    def inv_val(self) -> QPreTerm:
        return val_qpre(self._inv)

    @property
    def opt_pair_val(self) -> OptPairTerm:
        return val_opt_pair(self._opt_pair)
    
    @property
    def P_val(self) -> ProofHintTerm:
        return val_proof_hint(self._P)

    def prog_consistent(self, other: ProofHintTerm) -> bool:
        if isinstance(other, WhileHintTerm):
            return self._opt_pair == other._opt_pair\
                and self.P_val.prog_consistent(other.P_val)
        else:
            return False
    
    def str_content(self, prefix: str) -> str:
        r = prefix + "{ inv: " + self.inv_val.str_content() + "};\n"
        r += prefix + "while " + str(self._opt_pair) + " do\n"
        r += self.P_val.str_content(prefix + "\t") + "\n"
        r += prefix + "end"
        return r

class NondetHintTerm(ProofHintTerm):
    def __init__(self, proof_hints : Tuple[dts.Term, ...]):
        if not isinstance(proof_hints, tuple):
            raise ValueError()
        
        all_qvarls = QvarlsTerm(())
        for item in proof_hints:
            if not isinstance(item, dts.Term):
                raise ValueError()
            if item.type != type_proof_hint:
                raise RuntimeErrorWithLog("The term '" + str(item) + "' is not a proof hint.")
            item_val = val_proof_hint(item)
            all_qvarls = all_qvarls.join(item_val.all_qvarls)
        
        super().__init__(all_qvarls, "nondeterministic hint")
        self._proof_hints : Tuple[dts.Term,...] = proof_hints
    
    def get_proof_hint(self, i : int) -> ProofHintTerm:
        return val_proof_hint(self._proof_hints[i])

    def prog_consistent(self, other: ProofHintTerm) -> bool:
        if isinstance(other, NondetHintTerm):
            if len(self._proof_hints) != len(other._proof_hints):
                return False
            for i in range(len(self._proof_hints)):
                if not self.get_proof_hint(i).prog_consistent(other.get_proof_hint(i)):
                    return False
            return True
        else:
            return False

    def str_content(self, prefix: str) -> str:
        r = prefix + "(\n"
        r += self.get_proof_hint(0).str_content(prefix + "\t") + "\n"
        for i in range(1, len(self._proof_hints)):
            r += prefix + "#\n"
            r += self.get_proof_hint(i).str_content(prefix + "\t") + "\n"
        r += prefix + ")"
        return r

class SubproofHintTerm(ProofHintTerm):
    def __init__(self, subproof : dts.Term, arg_ls : dts.Term):
        # note: we need subproof to be a dts.Var
        if not isinstance(subproof, dts.Var) or not isinstance(arg_ls, dts.Term):
            raise ValueError()
        
        if subproof.type != type_proof:
            raise RuntimeErrorWithLog("The term '" + str(subproof) + "' is not a subproof.")
        
        arg_ls_val = val_qvarls(arg_ls)
        super().__init__(arg_ls_val, 'subproof hint')
        self._subproof : dts.Var = subproof
        self._arg_ls : dts.Term = arg_ls

    @property
    def subproof_val(self) -> ProofDefinedTerm:
        return val_proof(self._subproof)
    
    @property
    def arg_ls_val(self) -> QvarlsTerm:
        return val_qvarls(self._arg_ls)

    def prog_consistent(self, other: ProofHintTerm) -> bool:
        if isinstance(other, SubproofHintTerm):
            subproof_self = val_proof(self._subproof)
            subproof_other = val_proof(other._subproof)
            hint_self = val_proof_hint(subproof_self._proof_hint)
            hint_other = val_proof_hint(subproof_other._proof_hint)

            subprog_consistent = hint_self.prog_consistent(hint_other)\
                and subproof_self._arg_ls == subproof_other._arg_ls

            return subprog_consistent and self._arg_ls == other._arg_ls
        else:
            return False
    
    def str_content(self, prefix: str) -> str:
        # we need to reserve the variable name
        return prefix + str(self._subproof) + " " + str(self.arg_ls_val)
    
class QPreHintTerm(ProofHintTerm):
    def __init__(self, qpre : dts.Term):
        if not isinstance(qpre, dts.Term):
            raise ValueError()
        if qpre.type != type_qpre:
            raise ValueError()
        
        qpre_val = val_qpre(qpre)

        super().__init__(qpre_val.all_qvarls,"predicate hint")
        self._qpre : dts.Term = qpre_val

    @property
    def qpre_val(self) -> QPreTerm:
        return val_qpre(self._qpre)

    def prog_consistent(self, other: ProofHintTerm) -> bool:
        '''
        not meant for qpre
        '''
        raise Exception()
    
    def str_content(self, prefix: str) -> str:
        return prefix + str(self.qpre_val)


class UnionHintTerm(ProofHintTerm):
    def __init__(self, proof_hints : Tuple[dts.Term,...]):
        if not isinstance(proof_hints, tuple):
            raise ValueError()
        all_qvarls = QvarlsTerm(())
        for item in proof_hints:
            if not isinstance(item, dts.Term):
                raise ValueError()
            if item.type != type_proof_hint:
                raise RuntimeErrorWithLog("The term '" + str(item) + "' is not a proof hint.")
            item_val = val_proof_hint(item)
            all_qvarls = all_qvarls.join(item_val.all_qvarls)
        
        # check whether the program of all proofs are the same
        example_proof_hint = val_proof_hint(proof_hints[0])
        for i in range(1, len(proof_hints)):
            item_val = val_proof_hint(proof_hints[i])
            if not example_proof_hint.prog_consistent(item_val):
                raise RuntimeErrorWithLog(
                    "The (Union) rule requires that all the proofs are about the same program, but proof '" +\
                        str(proof_hints[0]) + "' and proof '" + str(proof_hints[i]) + "' are not."
                )


        super().__init__(all_qvarls, "union hint")
        self._proof_hints : Tuple[dts.Term,...] = proof_hints
    
    def get_proof_hint(self, i : int) -> ProofHintTerm:
        return val_proof_hint(self._proof_hints[i])

    def prog_consistent(self, other: ProofHintTerm) -> bool:
        if isinstance(other, UnionHintTerm):
            return self.get_proof_hint(0) == other.get_proof_hint(0)
        else:
            return False

    def str_content(self, prefix: str) -> str:
        r = prefix + "(\n"
        r += self.get_proof_hint(0).str_content(prefix + "\t") + "\n"
        for i in range(1, len(self._proof_hints)):
            r += prefix + ",\n"
            r += self.get_proof_hint(i).str_content(prefix + "\t") + "\n"
        r += prefix + ")"
        return r

class ProofSeqHintTerm(ProofHintTerm):
    def __init__(self, proof_hints : Tuple[dts.Term,...]):
        if not isinstance(proof_hints, tuple):
            raise ValueError()
        
        all_qvarls = QvarlsTerm(())
        for item in proof_hints:
            if not isinstance(item, dts.Term):
                raise ValueError()
            if item.type != type_proof_hint:
                raise RuntimeErrorWithLog("The term '" + str(item) + "' is not a proof hint.")
            # the individual subprogram can be "None" here
            item_val = val_proof_hint(item)
            all_qvarls = all_qvarls.join(item_val.all_qvarls)
        
        super().__init__(all_qvarls, "sequential hint")
        # flatten the sequential composition
        flattened = ()
        for item in proof_hints:
            if isinstance(item, ProofSeqHintTerm):
                flattened = flattened + item._proof_hints
            else:
                flattened = flattened + (item,)
        self._proof_hints : Tuple[dts.Term,...] = flattened

    def get_proof_hint(self, i : int) -> ProofHintTerm:
        return val_proof_hint(self._proof_hints[i])

    def prog_consistent(self, other: ProofHintTerm) -> bool:
        if isinstance(other, ProofSeqHintTerm):
            # get the list without qpredicate
            ls_self : List[ProofHintTerm] = []
            for i in range(len(self._proof_hints)):
                temp = self.get_proof_hint(i)
                if not isinstance(temp, QPreHintTerm):
                    ls_self.append(temp)
            ls_other : List[ProofHintTerm] = []
            for i in range(len(other._proof_hints)):
                temp = other.get_proof_hint(i)
                if not isinstance(temp, QPreHintTerm):
                    ls_other.append(temp)
            if len(ls_self) != len(ls_other):
                return False
            # begin comparing
            for i in range(len(ls_self)):
                if not ls_self[i].prog_consistent(ls_other[i]):
                    return False
            return True
        else:
            return False

    def str_content(self, prefix: str) -> str:
        if len(self._proof_hints) == 1:
            return self.get_proof_hint(0).str_content(prefix)
        elif len(self._proof_hints) > 1:
            r = ""
            for i in range(len(self._proof_hints)-1):
                r += self.get_proof_hint(i).str_content(prefix) + ";\n"
            r += self.get_proof_hint(len(self._proof_hints)-1).str_content(prefix)
            return r
        else:
            raise Exception()





'''
 Copyright 2022 Yingte Xu
 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at

 http://www.apache.org/licenses/LICENSE-2.0

 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.
'''

# ------------------------------------------------------------
# proof_term.py
#
# define the proof terms
# ------------------------------------------------------------

from __future__ import annotations
from typing import Any, List, Dict, Tuple

from nqpv import dts
from nqpv.vsystem.log_system import RuntimeErrorWithLog

from .qvarls_term import QvarlsTerm, type_qvarls, val_qvarls
from .opt_term import OperatorTerm
from .opt_pair_term import OptPairTerm, type_opt_pair, val_opt_pair
from . import qpre_term
from .qpre_term import QPreTerm, type_qpre, val_qpre
from .prog_term import AbortTerm, IfTerm, InitTerm, NondetTerm, ProgSttSeqTerm, ProgSttTerm, SkipTerm, UnitaryTerm, WhileTerm
from .scope_term import ScopeTerm

fac = dts.TermFact()
type_proof_stt = fac.axiom("proof_statement", fac.sort_term(0))
type_proof = fac.axiom("proof", fac.sort_term(0))



# proof statements

class ProofSttTerm(dts.Term):
    def __init__(self, pre : dts.Term, post : dts.Term):

        if not isinstance(pre, dts.Term) or not isinstance(post, dts.Term):
            raise ValueError()
        if pre.type != type_qpre:
            raise RuntimeErrorWithLog("The term '" + str(pre) + "' is not a quantum predicate.")
        if post.type != type_qpre:
            raise RuntimeErrorWithLog("The term '" + str(post) + "' is not a quantum predicate.")

        pre_val = val_qpre(pre)
        post_val = val_qpre(post)

        all_qvarls = pre_val.all_qvarls
        all_qvarls = all_qvarls.join(post_val.all_qvarls)
        super().__init__(type_proof_stt, None)
        self._all_qvarls : QvarlsTerm = all_qvarls
        self._pre : dts.Term = pre
        self._post : dts.Term = post
            
    @property
    def all_qvarls(self) -> QvarlsTerm:
        return self._all_qvarls
    
    @property
    def pre_val(self) -> QPreTerm:
        return val_qpre(self._pre)
    
    @property
    def post_val(self) -> QPreTerm:
        return val_qpre(self._post)

    def get_prog(self) -> ProgSttTerm:
        raise NotImplementedError()
    
    def str_content(self, prefix : str) -> str:
        raise NotImplementedError()

    def __str__(self) -> str:
        return "\n" + self.str_content("") + "\n"

    def expand(self, global_qvarls : set[str]) -> ProofSttTerm:
        '''
        get the proof statement with all subproofs substituted
        '''
        raise NotImplementedError()
    
def val_proof_stt(term : dts.Term) -> ProofSttTerm:
    if not isinstance(term, dts.Term):
        raise ValueError()
    if term.type != type_proof_stt:
        raise ValueError()
        
    if isinstance(term, ProofSttTerm):
        return term
    elif isinstance(term, dts.Var):
        val = term.val
        if not isinstance(val, ProofSttTerm):
            raise Exception()
        return val
    else:
        raise Exception()


# implement the weakest precondition calculus

class SkipProofTerm(ProofSttTerm):
    def __init__(self, pre : dts.Term, post : dts.Term):
        super().__init__(pre, post)
    
    def str_content(self, prefix: str) -> str:
        r = prefix + str(self.pre_val) + ";\n"
        r += prefix + "skip"
        return r
    

class AbortProofTerm(ProofSttTerm):
    def __init__(self, pre : dts.Term, post : dts.Term):
        super().__init__(pre, post)
    
    def str_content(self, prefix: str) -> str:
        r = prefix + str(self.pre_val) + ";\n"
        r += prefix + "abort"
        return r
    

class InitProofTerm(ProofSttTerm):
    def __init__(self, pre : dts.Term, post : dts.Term, qvarls : dts.Term):
        if not isinstance(qvarls, dts.Term):
            raise ValueError()
        if qvarls.type != type_qvarls:
            raise ValueError()

        super().__init__(pre, post)
        self._all_qvarls = self._all_qvarls.join(val_qvarls(qvarls))
        self._qvarls : dts.Term = qvarls
    
    @property
    def qvarls_val(self) -> QvarlsTerm:
        return val_qvarls(self._qvarls)
    
    def str_content(self, prefix: str) -> str:
        r = prefix + str(self.pre_val) + ";\n"
        r += prefix + str(self.qvarls_val) + " :=0"
        return r
        
class UnitaryProofTerm(ProofSttTerm):
    def __init__(self, pre : dts.Term, post : dts.Term, opt_pair : dts.Term):
        if not isinstance(opt_pair, dts.Term):
            raise ValueError()
        if opt_pair.type != type_opt_pair:
            raise ValueError()

        super().__init__(pre, post)
        self._all_qvarls = self._all_qvarls.join(val_opt_pair(opt_pair).qvarls_val)
        self._opt_pair : dts.Term = opt_pair
    
    @property
    def opt_pair_val(self) -> OptPairTerm:
        return val_opt_pair(self._opt_pair)
    
    def str_content(self, prefix: str) -> str:
        r = prefix + str(self.pre_val) + ";\n"
        r += prefix + str(self.opt_pair_val.qvarls_val) + " *= " + str(self.opt_pair_val.opt)
        return r
        
class IfProofTerm(ProofSttTerm):
    def __init__(self, pre : dts.Term, post : dts.Term, opt_pair : dts.Term, P1 : dts.Term, P0 : dts.Term):        
        if not isinstance(opt_pair, dts.Term) or not isinstance(P1, dts.Term) or not isinstance(P0, dts.Term):
            raise ValueError()
        if opt_pair.type != type_opt_pair:
            raise ValueError()
        if P1.type != type_proof_stt:
            raise RuntimeErrorWithLog("The term '" + str(P1) + "' is not a proof statement.")
        if P0.type != type_proof_stt:
            raise RuntimeErrorWithLog("The term '" + str(P0) + "' is not a proof statement.")

        super().__init__(pre, post)
        self._all_qvarls = self._all_qvarls.join(val_opt_pair(opt_pair).qvarls_val)
        self._all_qvarls = self._all_qvarls.join(val_proof_stt(P1).all_qvarls)
        self._all_qvarls = self._all_qvarls.join(val_proof_stt(P0).all_qvarls)
        self._opt_pair = opt_pair
        self._P0 = P0
        self._P1 = P1
    
    @property
    def opt_pair_val(self) -> OptPairTerm:
        return val_opt_pair(self._opt_pair)
    
    @property
    def P1_val(self) -> ProofSttTerm:
        return val_proof_stt(self._P1)
    
    @property
    def P0_val(self) -> ProofSttTerm:
        return val_proof_stt(self._P0)
    
    def str_content(self, prefix: str) -> str:
        r = prefix + str(self.pre_val) + ";\n"
        r += prefix + "if " + str(self.opt_pair_val) + " then\n"
        r += self.P1_val.str_content(prefix + "\t") + "\n"
        r += prefix + "else\n"
        r += self.P0_val.str_content(prefix + "\t") + "\n"
        r += prefix + "end"
        return r
    
    
class WhileProofTerm(ProofSttTerm):
    def __init__(self, pre : dts.Term, post : dts.Term, inv : dts.Term, opt_pair : dts.Term, P : dts.Term):        
        if not isinstance(inv, dts.Term) or not isinstance(opt_pair, dts.Term) or not isinstance(P, dts.Term):
            raise ValueError()
        if inv.type != type_qpre or opt_pair.type != type_opt_pair:
            raise ValueError()
        if P.type != type_proof_stt:
            raise RuntimeErrorWithLog("The term '" + str(P) + "' is not a proof statement.")

        super().__init__(pre, post)
        self._all_qvarls = self._all_qvarls.join(val_qpre(inv).all_qvarls)
        self._all_qvarls = self._all_qvarls.join(val_opt_pair(opt_pair).qvarls_val)
        self._all_qvarls = self._all_qvarls.join(val_proof_stt(P).all_qvarls)
        self._inv = inv
        self._opt_pair = opt_pair
        self._P = P
    
    @property
    def inv_val(self) -> QPreTerm:
        return val_qpre(self._inv)

    @property
    def opt_pair_val(self) -> OptPairTerm:
        return val_opt_pair(self._opt_pair)
    
    @property
    def P_val(self) -> ProofSttTerm:
        return val_proof_stt(self._P)
    
    def str_content(self, prefix: str) -> str:
        r = prefix + str(self.pre_val) + ";\n"
        r += prefix + "{ inv: " + self.inv_val.str_content() + " };\n"
        r += prefix + "while " + str(self.opt_pair_val) + " do\n"
        r += self.P_val.str_content(prefix + "\t") + "\n"
        r += prefix + "end"
        return r
    
class NondetProofTerm(ProofSttTerm):
    def __init__(self, pre : dts.Term, post : dts.Term, proof_ls : Tuple[dts.Term,...]):        
        if not isinstance(proof_ls, tuple):
            raise ValueError()
        for item in proof_ls:
            item_val = val_proof_stt(item)
        
        super().__init__(pre, post)
        for item in proof_ls:
            item_val = val_proof_stt(item)
            self._all_qvarls = self._all_qvarls.join(item_val.all_qvarls)
        self._proof_ls : Tuple[dts.Term,...] = proof_ls
    
    def get_proof(self, i : int) -> ProofSttTerm:
        return val_proof_stt(self._proof_ls[i])
    
    def str_content(self, prefix: str) -> str:
        r = prefix + str(self.pre_val) + ";\n"
        r += prefix + "(\n"
        r += self.get_proof(0).str_content(prefix + "\t") + "\n"
        for i in range(1, len(self._proof_ls)):
            r += prefix + "#\n"
            r += self.get_proof(i).str_content(prefix + "\t") + "\n"
        r += prefix + ")"
        return r
    
class SubproofTerm(ProofSttTerm):
    def __init__(self, pre : dts.Term, post : dts.Term, subproof : dts.Term, arg_ls : dts.Term):
        if not isinstance(subproof, dts.Term) or not isinstance(arg_ls, dts.Term):
            raise ValueError()
        if subproof.type != type_proof or arg_ls.type != type_qvarls:
            raise ValueError()

        super().__init__(pre, post)
        self._all_qvarls = val_qvarls(arg_ls)
        self._subproof = subproof
        self._arg_ls = arg_ls
    
    @property
    def arg_ls_val(self) -> QvarlsTerm:
        return val_qvarls(self._arg_ls)
    
    def str_content(self, prefix: str) -> str:
        r = prefix + str(self._pre) + ";\n"
        r += prefix + str(self._subproof) + " " + str(self.arg_ls_val)
        return r
    
class QPreProofTerm(ProofSttTerm):
    def __init__(self, pre : dts.Term, post : dts.Term, qpre : dts.Term):
        if not isinstance(qpre, dts.Term):
            raise ValueError()
        if qpre.type != type_qpre:
            raise ValueError()
        
        super().__init__(pre, post)
        self._all_qvarls = self._all_qvarls.join(val_qpre(qpre).all_qvarls)
        self._qpre = qpre
    
    @property
    def qpre_val(self) -> QPreTerm:
        return val_qpre(self._qpre)
    
    def str_content(self, prefix: str) -> str:
        return prefix + str(self.qpre_val)
    
class UnionProofTerm(ProofSttTerm):
    def __init__(self, pre : dts.Term, post : dts.Term, proof_ls : Tuple[dts.Term,...]):
        if not isinstance(proof_ls, tuple):
            raise ValueError()
        for item in proof_ls:
            item_val = val_proof_stt(item)

        super().__init__(pre, post)
        for item in proof_ls:
            item_val = val_proof_stt(item)
            self._all_qvarls = self._all_qvarls.join(item_val.all_qvarls)
        self._proof_ls : Tuple[dts.Term,...] = proof_ls
    
    def get_proof(self, i : int) -> ProofSttTerm:
        return val_proof_stt(self._proof_ls[i])
    
    def str_content(self, prefix: str) -> str:
        r = prefix + str(self.pre_val) + ";\n"
        r += prefix + "(\n"
        r += self.get_proof(0).str_content(prefix + "\t") + ";\n"
        r += prefix + "\t" + str(self.get_proof(0).post_val) + "\n"
        for i in range(1, len(self._proof_ls)):
            r += prefix + ",\n"
            r += self.get_proof(i).str_content(prefix + "\t") + ";\n"
            r += prefix + "\t" + str(self.get_proof(i).post_val) + "\n"
        r += prefix + ")"
        return r
    

class ProofSeqTerm(ProofSttTerm):
    def __init__(self, pre : dts.Term, post : dts.Term, proof_ls : Tuple[dts.Term,...]):        
        if not isinstance(proof_ls, tuple):
            raise ValueError()
        for item in proof_ls:
            item_val = val_proof_stt(item)
        
        super().__init__(pre, post)
        for item in proof_ls:
            item_val = val_proof_stt(item)
            self._all_qvarls = self._all_qvarls.join(item_val.all_qvarls)
        self._proof_ls : Tuple[dts.Term,...] = proof_ls

    def get_proof(self, i : int) -> ProofSttTerm:
        return val_proof_stt(self._proof_ls[i])

    def str_content(self, prefix: str) -> str:
        if len(self._proof_ls) == 1:
            return self.get_proof(0).str_content(prefix)
        elif len(self._proof_ls) > 1:
            r = ""
            for i in range(len(self._proof_ls)-1):
                r += self.get_proof(i).str_content(prefix) + ";\n\n"
            r += self.get_proof(len(self._proof_ls)-1).str_content(prefix)
            return r
        else:
            raise Exception()
        

class ProofTerm(dts.Term):
    def __init__(self, arg_ls : dts.Term):
        if arg_ls.type != type_qvarls:
            raise RuntimeErrorWithLog("The term '" + str(arg_ls) + "' is not a quantum variable list.")
        arg_ls_val = val_qvarls(arg_ls)

        super().__init__(type_proof, None)
        self._arg_ls : dts.Term = arg_ls
        self._all_qvarls : QvarlsTerm = arg_ls_val

    @property
    def all_qvarls(self) -> QvarlsTerm:
        return self._all_qvarls

    @property
    def arg_ls_val(self) -> QvarlsTerm:
        return val_qvarls(self._arg_ls)

    def __eq__(self, other) -> bool:
        return NotImplemented

    def __str__(self) -> str:
        raise NotImplementedError()

class ProofDefiningTerm(ProofTerm):
    '''
    The proof being defined.
    '''
    def __str__(self) -> str:
        return "\n(Proof Being Defined) " + str(self.arg_ls_val) + "\n"

class ProofDefinedTerm(ProofTerm):
    '''
    the completed proof
    '''
    def __init__(self, pre : dts.Term, proof_hint : dts.Term, proof_stts : dts.Term, post : dts.Term, arg_ls : dts.Term):
        '''
        proof_hint : proof_hint is necessary for apply_hint method
        the specified pre and post conditions are necessary, because they are not the full extension
        '''
        
        if not isinstance(pre, dts.Term)\
             or not isinstance(proof_hint, dts.Term)\
             or not isinstance(proof_stts, dts.Term)\
             or not isinstance(post, dts.Term)\
             or not isinstance(arg_ls, dts.Term):
            raise ValueError()

        super().__init__(arg_ls)
        self._arg_ls : dts.Term = arg_ls
        self._proof_hint : dts.Term = proof_hint
        self._proof_stts : dts.Term = proof_stts
        self._pre = pre
        self._post = post
        self._all_qvarls = self._all_qvarls.join(val_proof_stt(self._proof_stts).all_qvarls)

    
    @property
    def proof_stts_val(self) -> ProofSttTerm:
        return val_proof_stt(self._proof_stts)

    @property
    def pre_val(self) -> QPreTerm:
        return val_qpre(self._pre)
    
    @property
    def post_val(self) -> QPreTerm:
        return val_qpre(self._post)
        

    def __str__(self) -> str:
        r = "\nproof " + str(self.arg_ls_val) + " : \n" 
        r += "\t" + str(self.pre_val) + ";\n\n"
        r += self.proof_stts_val.str_content("\t") + ";\n"
        r += "\n\t" + str(self.post_val) + "\n"
        return r

def val_proof(term : dts.Term) -> ProofDefinedTerm:
    if not isinstance(term, dts.Term):
        raise ValueError()
    if term.type != type_proof:
        raise ValueError()
        
    if isinstance(term, ProofDefinedTerm):
        return term
    elif isinstance(term, dts.Var):
        val = term.val
        if not isinstance(val, ProofDefinedTerm):
            raise Exception()
        return val
    else:
        raise Exception()



'''
 Copyright 2022 Yingte Xu
 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at

 http://www.apache.org/licenses/LICENSE-2.0

 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.
'''

# ------------------------------------------------------------
# logsystem.py
#
# Provide the error/information log and report system.
# This system shall be used by all modules in this project to store informations.
# Other modules will transform information through different channels in LogSystem.channel_dict.
# ------------------------------------------------------------

from __future__ import annotations
from typing import Any, List, Dict
from io import TextIOWrapper

from .syntax.pos_info import PosInfo

class LogSystem:

    channels : Dict[str, LogSystem] = {}

    def __new__(cls, name : str, prefix : str = "", end : str = "\n"):
        '''
        single case design
        '''
        if not isinstance(name, str) or not isinstance(prefix, str) or not isinstance(end, str):
            raise ValueError()

        if name not in LogSystem.channels:
            channel = super().__new__(cls)

            # the list to save all kinds of logs (not only string)
            channel.logs = []
            # how to begin an item when output
            channel.prefix = prefix
            # how to end an item when output
            channel.end = end

            LogSystem.channels[name] = channel

        return LogSystem.channels[name]
        

    def __init__(self, name : str, prefix : str = "", end : str = "\n"):
        '''
        create a new channel

        name: channel name
        '''
        self.logs : List[Any]
        self.prefix : str
        self.end : str
        

    @property
    def empty(self) -> bool:
        return len(self.logs) == 0

    def append(self, data : Any) -> None:
        self.logs.append(data)

    def get_front(self, pfile : None | TextIOWrapper = None, cmd_print : bool = False, drop : bool = True) -> str:
        if drop:
            result : str =  self.prefix + str(self.logs.pop(0)) + self.end
        else:
            result : str = self.prefix + self.logs[0] + self.end
        
        if pfile is not None:
            pfile.write(result)

        if cmd_print:
            print(result, end = "")

        return result

    def get_back(self, pfile : None | TextIOWrapper = None, cmd_print : bool = False, drop : bool = True) -> str:
        if drop:
            result : str =  self.prefix + str(self.logs.pop(len(self.logs)-1)) + self.end
        else:
            result : str = self.prefix + self.logs[len(self.logs)-1] + self.end
        
        if pfile is not None:
            pfile.write(result)

        if cmd_print:
            print(result, end = "")

        return result

    
    def summary(self, pfile : None | TextIOWrapper = None, cmd_print : bool = False, drop : bool = True) -> str:
        '''
        Conclude and return all the informations. Clean self.logs afterwards.
        pfile: the file pointer to write this summary
        '''

        result : str = ""
        for info in self.logs:
            result += self.prefix + str(info) + self.end

        if drop:
            self.logs.clear()
        
        if pfile is not None:
            pfile.write(result)

        if cmd_print:
            print(result, end = "")

        return result

    def single(self, data : Any, pfile : None | TextIOWrapper = None, cmd_print : bool = True) -> str:
        '''
        push in the information, polish it, pop out and write to file/print out immediately.
        '''
        self.append(data)
        return self.get_back(pfile, cmd_print, True)
    
    def __len__(self) -> int:
        return len(self.logs)


class RuntimeErrorWithLog(RuntimeError):
    '''
    this error class automatically logs the error information.
    '''
    def __init__(self, msg : str, pos : PosInfo | None = None):
        super().__init__(msg, pos)
        LogSystem.channels["error"].append(msg + PosInfo.str(pos))


'''
 Copyright 2022 Yingte Xu
 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at

 http://www.apache.org/licenses/LICENSE-2.0

 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.
'''

# ------------------------------------------------------------
# vlexer.py
#
# tokenizer
# ------------------------------------------------------------

from __future__ import annotations
from typing import Any, List

import ply.lex as lex

from nqpv.vsystem.log_system import LogSystem

from .pos_info import PosInfo


reserved = {
    # for vsystem
    'def'   : 'DEF',
    'example'   : 'EXAMPLE',
    'axiom' : 'AXIOM',
    'setting'   : 'SETTING',
    'show'  : 'SHOW',
    'save'  : 'SAVE',
    'at'    : 'AT',

    # constants
    'true'  : "TRUE",
    "false" : "FALSE",

    # settings
    'EPS'   : 'EPS',
    'SDP_PRECISION' : 'SDP_PRECISION',
    'SILENT'    : 'SILENT',

    # inner calculation methods
    'import': 'IMPORT',
    'load'  : 'LOAD',

    # keywords for type
    'operator'  : 'OPERATOR',
    'scope' : 'SCOPE',
    'program'   : 'PROGRAM',
    'proof' : 'PROOF',

    # for programs and verifications
    'skip'  : 'SKIP',
    'abort' : 'ABORT',
    'if'    : 'IF',
    'then'  : 'THEN',
    'else'  : 'ELSE',
    'while' : 'WHILE',
    'do'    : 'DO',
    'end'   : 'END',
    'inv'   : 'INV',
}

# List of token names.
tokens = [
    'ID',
    'FLOAT_NUM',
    'STRING',
    'INIT',
    'ASSIGN',
    'MUL_EQ',
 ] + list(reserved.values())

# Regular expression rules for simple tokens
t_INIT = r':=0'
t_ASSIGN = r':='
t_MUL_EQ = r'\*='

literals = ['.', ',', ';', '#', ':', '[', ']', '(', ')', '{', '}']


def t_STRING(t):
    r'".*"'
    return t

# use // or /* */ to comment
def t_COMMENT(t):
    r'(/\*(.|\n)*?\*/)|(//.*)'
    for c in t.value:
        if c == '\n':
            t.lexer.lineno += 1

def t_ID(t):
    r'[a-zA-Z_][a-zA-Z_0-9]*'
    t.type = reserved.get(t.value,'ID')    # Check for reserved words
    return t

def t_FLOAT_NUM(t):
    r'(\+|-)?([1-9]\d*|0)(\.\d+)?([Ee](\+|-)?\d+)?'
    try:
        # test the transform
        float(t.value)
        return t
    except:
        LogSystem.channels["error"].append("Syntax Error. Illegal number '" + t.value[0] + "'." + str(PosInfo(t.lineno)))
        t.lexer.skip(1)

# Define a rule so we can track line numbers
def t_newline(t):
    r'\n+'
    t.lexer.lineno += len(t.value)


# A string containing ignored characters (spaces and tabs)
t_ignore = ' \t'


def t_error(t):
    LogSystem.channels["error"].append("Syntax Error. Illegal character '" + t.value[0] + "'." + str(PosInfo(t.lineno)))
    t.lexer.skip(1)


# Build the lexer
lexer = lex.lex()


'''
 Copyright 2022 Yingte Xu
 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at

 http://www.apache.org/licenses/LICENSE-2.0

 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.
'''

# Dependent Type System Support
# To extend 'terms', inherbit from Term, and overload the following methods:
#  __eq__, substitute, eval

from __future__ import annotations
from typing import List, Tuple, Dict


# No polymorphism on different universe

class DTS_RuntimeError(RuntimeError):
    def __init__(self, msg : str):
        super().__init__(msg)

class Para:
    '''
    paremeter for functions
    Readonly. No copy needed in assignment.    
    '''
    def __init__(self, para_id : str, type : Term):
        self._para_id : str = para_id
        self._para_type : Term = type
    
    @property
    def id(self) -> str:
        return self._para_id

    @property
    def type(self) -> Term:
        return self._para_type

    @property
    def is_type(self) -> bool:
        '''
        decide whether the term is also a type
        '''
        return isinstance(self.type, Sort)

    def __str__(self) -> str:
        return self.id
    
    def __eq__(self, other) -> bool:
        if isinstance(other, Para):
            # since id is unique
            return self.id == other.id
        else:
            raise NotImplemented

    def set_parent_env(self, parent : ParaMicroEnv) -> None:
        self.type.set_parent_env(parent)

    def substitute(self, para : Para, val : Term | Para) -> Para | Term:
        if self == para:
            return val

        return Para(self._para_id, self._para_type.substitute(para, val))

class ParaMicroEnv:
    '''
    The micro (but still inductive) environment for parameters in functions
    Micro means at every level only one parameter is stored.
    Note: naming overload is NOT allowed, and the users should pay attention
    '''

    def __init__(self, para : Para | None = None):
        '''
        para : None means no parameter defined in this environment
        '''
        self._parent_env : ParaMicroEnv | None = None
        self._para : Para | None = para
    
    @property
    def para(self) -> Para | None:
        return self._para
    
    @property
    def parent(self) -> ParaMicroEnv | None:
        '''
        should not be modified
        '''
        return self._parent_env

    def set_parent(self, parent : ParaMicroEnv | None) -> None:
        if parent is None:
            return
        if isinstance(parent, ParaMicroEnv):
            if self.parent is not None:
                raise DTS_RuntimeError("The parent environment has already been set.")
            # set the parent environment
            self._parent_env = parent
        else:
            raise Exception()

    def get_term(self, ask_id : str) -> Term:
        '''
        Get the type term with the ask_id.
        ask_id: should be a string
        '''
        if self._para is not None:
            if self._para.id == ask_id:
                return self._para.type
        if self._parent_env is not None:
            return self._parent_env.get_term(ask_id)
        else:
            raise DTS_RuntimeError("Variable of id '" + ask_id + "' does not exist.")
    
    def __str__(self) -> str:
        return str(self._para)

class Term:
    '''
    Readonly. No copy needed in assignment.
    '''
    def __init__(self, type : Term | None, para_env : ParaMicroEnv | None):
        self._type : Term | None = type
        self._para_env : ParaMicroEnv | None = para_env
    
    @property
    def type(self) -> Term:
        if self._type is None:
            raise DTS_RuntimeError("Maximum universe number reached.")
        return self._type
    
    @property
    def para_env(self) -> ParaMicroEnv | None:
        return self._para_env

    def set_parent_env(self, parent : ParaMicroEnv) -> None:
        if self._para_env is not None:
            self._para_env.set_parent(parent)

    @property
    def is_type(self) -> bool:
        '''
        decide whether the term is also a type
        '''
        return isinstance(self.type, Sort)
            
    def substitute(self, para : Para,  val : Term | Para) -> Term:
        '''
        return the new term with the parameter substituted
        WARNING: no type check for val is applied
        '''        

        return self

    def eval(self) -> Term:
        return self

    def __eq__(self, other) -> bool:
        return NotImplemented

class Sort(Term):
    '''
    Readonly. No copy needed in assignment.
    '''
    def __init__(self, type : Sort | None, u_num : int):
        super().__init__(type, None)
        self._u_num : int = u_num

    def __str__(self) -> str:
        return "Sort " + str(self._u_num)

    @property
    def u_num(self) -> int:
        return self._u_num
    
    def __eq__(self, other : Sort) -> bool:
        # sorts only equal to sorts
        if isinstance(other, Sort):
            return self.u_num == other.u_num
        else:
            return False

class Axiom(Term):
    '''
    The axiom does not have an value, i.e. it does not depend on any value
    Readonly. No copy needed in assignment.
    '''
    def __init__(self, axiom_id : str, type : Term):
        super().__init__(type, None)
        self._axiom_id : str = axiom_id

    @property
    def id(self) -> str:
        return self._axiom_id

    def __str__(self) -> str:
        return self._axiom_id

    def __eq__(self, other : Axiom) -> bool:
        if isinstance(other, Axiom):
            return self.id == other.id
        else:
            return False

class Var(Term):
    '''
    variable with identifier and value.
    Readonly. No copy needed in assignment.
    '''
    def __init__(self, var_id : str, type : Term, value : Term, var_name : str = ""):
        '''
        the variable value 'value' must be of the type 'type'
        '''
        super().__init__(type, None)
        self._var_id : str = var_id
        # ensure that the value is not a variable
        if isinstance(value, Var):
            self._value : Term = value._value
        else:
            self._value : Term = value

        # name decides how this variable is displayed
        # note that different variable have different ids, but can have the same name
        self._var_name : str = var_name
    
    @property
    def id(self) -> str:
        return self._var_id
    
    @property
    def name(self) -> str:
        return self._var_name
    
    @property
    def val(self) -> Term:
        return self._value

    def __str__(self) -> str:
        if self._var_name != "":
            return self._var_name
        else:
            return self.id
    
    def __eq__(self, other : Term) -> bool:
        if isinstance(other, Var):
            if self.id == other.id:
                return True
            elif self.val == other.val:
                return True
            else:
                return False
        else:
            return self.val == other
    
    def eval(self) -> Term:
        return self._value.eval()


class ToType(Term):
    '''
    Construct a term as a function type
    Readonly. No copy needed in assignment.
    '''
    def __init__(self, type : Term, para : Para, type_right : Term | Para):
        '''
        Construction of term: [para_env] -> type_right
        The 'para_env' should point to the ParaMicroEnv which contains the parameter for this function
        '''
        local_env = ParaMicroEnv(para)
        para.type.set_parent_env(local_env)
        type_right.set_parent_env(local_env)

        super().__init__(type, local_env)

        self._para_env : ParaMicroEnv = self._para_env
        self._type_right : Term | Para = type_right

    @property
    def para_env(self) -> ParaMicroEnv:
        return self._para_env
    
    @property
    def para(self) -> Para:
        if self._para_env.para is None:
            raise Exception()
        return self._para_env.para
    
    @property
    def type_right(self) -> Term | Para:
        return self._type_right

    def __str__(self) -> str:
        return "(" + str(self.para) + " : " + str(self.para.type) + ") -> " + str(self._type_right)
        
    
    def __eq__(self, other) -> bool:
        '''
        we require that the equivalence check will consider different parameter naming
        '''
        if isinstance(other, ToType):

            '''
            check whether the two parameters paraA and paraB are the same, with respect to the two environments
            '''
            type_eq = self.para.type == other.para.type

            return type_eq and self.type_right == other.type_right

        else:
            raise NotImplemented

    def substitute(self, para: Para, val: Term | Para) -> Term:
        if self._type is None:
            raise Exception()
        else:
            new_type = self._type.substitute(para, val)

        new_para = self.para.substitute(para, val)
        if isinstance(new_para, Term):
            # this should not happen, because parameter id are unique
            raise Exception()

        new_term = ToType(new_type, new_para, self._type_right.substitute(para, val))
        new_term.set_parent_env(self._para_env)
        return new_term        


class Lambda(Term):
    '''
    Construct a term as an anonymous function
    '''
    def __init__(self, type : Term, input : Para, output : Term | Para):
        local_env = ParaMicroEnv(input)
        input.type.set_parent_env(local_env)
        output.set_parent_env(local_env)

        super().__init__(type, local_env)
        self._para_env : ParaMicroEnv = self._para_env
        self._output : Term | Para = output
    
    @property
    def para_env(self) -> ParaMicroEnv:
        return self._para_env

    @property
    def para(self) -> Para:
        if self._para_env.para is None:
            raise Exception()
        return self._para_env.para

    def eval_on_val(self, val : Term | Para) -> Term:
        '''
        evaluate the output using 'val' as input
        WARNING : no type check
        '''

        result = self._output.substitute(self.para, val)
        if not isinstance(result, Term):
            raise Exception()
        return result
    
    def __str__(self) -> str:
        return "fun(" + str(self.para) + " : " + str(self.para.type) + ") => " + str(self._output)
        
    def __eq__(self, other) -> bool:
        if isinstance(other, ToType):
            return False
        else:
            raise NotImplemented

    def substitute(self, para: Para, val: Term | Para) -> Term:
        if self._type is None:
            raise Exception()
        else:
            new_type = self._type.substitute(para, val)

        new_para = self.para.substitute(para, val)
        if isinstance(new_para, Term):
            # this should not happen, because parameter id are unique
            raise Exception()

        new_term = Lambda(new_type, new_para, self._output.substitute(para, val))
        return new_term


class Apply(Term):
    '''
    Construct a term as a function application
    Readonly. No copy needed in assignment.
    '''
    def __init__(self, type : Term, fun : Term, para : Term | Para):
        '''
        the type of fun should be a 'function'
        '''
        local_env = ParaMicroEnv()
        para.set_parent_env(local_env)
        fun.set_parent_env(local_env)

        super().__init__(type, local_env)

        self._para_env : ParaMicroEnv
        self._term_fun : Term = fun
        self._term_val : Term | Para = para

    @property
    def para_env(self) -> ParaMicroEnv:
        return self._para_env

    def __str__(self) -> str:
        return "(" + str(self._term_fun) + " " + str(self._term_val) + ")"

    def eval(self) -> Term:
        '''
        evaluate the application and return the result
        '''

        if self.para_env.parent is not None:
            raise DTS_RuntimeError("invalid function evaluation")
        
        elif isinstance(self._term_fun, Axiom):
            raise DTS_RuntimeError("Axiom '" + str(self._term_fun) + "' is not executable.")

        elif isinstance(self._term_fun, Var):
            # apply the substitution
            temp = Apply(self.type, self._term_fun.val, self._term_val)
            result = temp.eval()
            # set the parent environment to None
            result.set_parent_env(ParaMicroEnv())
            return result
        
        elif isinstance(self._term_fun, Lambda):
            result = self._term_fun.eval_on_val(self._term_val)
            # set the parent environment to None
            result.set_parent_env(ParaMicroEnv())
            return result
        
        else:
            raise Exception()

    def __eq__(self, other) -> bool:
        return self.eval() == other

    def substitute(self, para: Para, val: Term | Para) -> Term:
        if self._type is None:
            raise Exception()
        else:
            new_type = self._type.substitute(para, val)

        new_term_fun = self._term_fun.substitute(para, val)
        if not isinstance(new_term_fun, ToType):
            raise Exception()

        new_term = Apply(new_type, new_term_fun, self._term_val.substitute(para, val))
        new_term.set_parent_env(self._para_env)
        return new_term
        


class TermFact:
    '''
    Apply the factory model. New instance checks are done here.
    '''

    __instance : TermFact | None = None
    __max_sort_level : int = 16

    def __new__(cls):
        if TermFact.__instance is not None:
            return TermFact.__instance

        instance = super().__new__(cls)
        #initialize a term factory

        if TermFact.__max_sort_level < 0:
            raise DTS_RuntimeError("Invalid maximum sort level.")

        # the list of all sorts, the index corresponds to the universe number
        instance._sort_list = [Sort(None, TermFact.__max_sort_level)]

        for u_num in range(TermFact.__max_sort_level-1, -1, -1):
            upper_sort = instance._sort_list[0]
            instance._sort_list.insert(0, Sort(upper_sort, u_num))
        
        TermFact.__instance = instance
        return instance

    def __init__(self):
        self._sort_list : List[Sort]

    # term construction methods
    def sort_term(self, u_num : int) -> Sort:
        return self._sort_list[u_num]

    def axiom(self, axiom_id : str, type : Term) -> Axiom:
        if not isinstance(axiom_id, str):
            raise DTS_RuntimeError("invalid axiom id")
        if not isinstance(type, Term):
            raise DTS_RuntimeError("invalid type")
        if not type.is_type:
            raise DTS_RuntimeError("The term '" + str(type) + "' is not a type.")
        
        return Axiom(axiom_id, type)

    def var(self, var_id : str, type : Term, value : Term) -> Var:
        '''
        variable should not contain free parameters
        '''
        if not isinstance(var_id, str):
            raise DTS_RuntimeError("invalid variable id")
        if not isinstance(type, Term):
            raise DTS_RuntimeError("invalid type")
        if not type.is_type:
            raise DTS_RuntimeError("The term '" + str(type) + "' is not a type.")
        if not isinstance(value, Term):
            raise DTS_RuntimeError("invalid value")
        # check whether the type and value corresponds
        if not value.type == type:
            raise DTS_RuntimeError("The value '" + str(value) + "' is not of type '" + str(type) + "'.")
        
        self.no_para_check(type)
        self.no_para_check(value)

        return Var(var_id, type, value)
    
    def _no_para_check_iter(self, term : Term | Para, para_env : ParaMicroEnv) -> None:
        if isinstance(term, Axiom) or isinstance(term, Sort) or isinstance(term, Var):
            return
        elif isinstance(term, Para):
            try:
                para_env.get_term(term.id)
                return
            except DTS_RuntimeError:
                raise DTS_RuntimeError("free parameter detected")
        elif isinstance(term, ToType):
            self._no_para_check_iter(term.para.type, para_env)
            self._no_para_check_iter(term.type, term.para_env)
            return
        elif isinstance(term, Lambda):
            self._no_para_check_iter(term.para.type, para_env)
            self._no_para_check_iter(term._output, term.para_env)
            return
        elif isinstance(term, Apply):
            self._no_para_check_iter(term._term_fun, term.para_env)
            self._no_para_check_iter(term._term_val, term.para_env)
            return
            
        raise Exception()


    def no_para_check(self, term : Term) -> None:
        '''
        check that there is no free parameter in this term
        '''
        return self._no_para_check_iter(term, ParaMicroEnv())
    
    def _fun_con_type(self, type1 : Term | Para, type2 : Term | Para) -> Sort:
        '''
        get the type of function construction (a : type1) -> type2
        '''
        if not isinstance(type1.type, Sort) or not isinstance(type2.type, Sort):
            raise DTS_RuntimeError("type1 and type2 should be types")
        
        return self._sort_list[max(type1.type.u_num, type2.type.u_num)]

    def para(self,  para_id : str, type : Term) -> Para:
        if not isinstance(para_id, str):
            raise DTS_RuntimeError("invalid parameter id")
        if not isinstance(type, Term):
            raise DTS_RuntimeError("invalid type")
        if not type.is_type:
            raise DTS_RuntimeError("The term '" + str(type) + "' is not a type.")
        return Para(para_id, type)

    def to_type(self, para : Para, type_right : Term | Para) -> ToType:
        if not isinstance(para, Para):
            raise DTS_RuntimeError("invalid parameter")
        if not isinstance(type_right, Term) and not isinstance(type_right, Para):
            raise DTS_RuntimeError("invalid right type")
        if not type_right.is_type:
            raise DTS_RuntimeError("The right type term '" + str(type_right) + "' is not a type.")

        if para is None:
            raise DTS_RuntimeError("The environment should not be empty")

        # evaluate the type
        type = self._fun_con_type(para.type, type_right)

        # copy the para_env to avoid unexpected manipulations
        return ToType(type, para, type_right)    

    def _lambda_type(self, input : Para, output: Term) -> Term:
        '''
        get the type of a lambda expression
        '''
        return self.to_type(input, output.type)

    def fun_lambda(self, input : Para, output : Term | Para) -> Lambda:
        if not isinstance(input, Para):
            raise DTS_RuntimeError("invalid lambda parameter")
        if not isinstance(output, Term) and not isinstance(output, Para):
            raise DTS_RuntimeError("invalid output for lambda expression")

        return Lambda(self.to_type(input, output.type), input, output)


    def fun_apply(self, fun : Term, value : Term | Para) -> Apply:
        if not isinstance(fun.type, ToType):
            raise DTS_RuntimeError("The term '" + str(fun) + "' is not a function." )
        
        if not isinstance(value, Term) and not isinstance(value, Para):
            raise DTS_RuntimeError("The value parameter to apply is neither a parameter nor a term.")

        type_needed = fun.type.para.type
        if not type_needed == value.type:
            raise DTS_RuntimeError("The value parameter '" + str(value) + "' of type '" + str(fun.type.para.type) + 
                "cannot be applied here, where a value of type '" + str(type_needed) + "' is needed.")
        
        # calculate the type
        apply_type = fun.type.type_right.substitute(fun.type.para, value)
        if isinstance(apply_type, Para):
            raise Exception()
        
        return Apply(apply_type, fun, value)

    def fun_evl(self, fun : ToType, value : Term) -> Term:
        '''
        substitue the parameter, evaluate and return the result
        the evaluation happens at the fun_apply structures
        '''
        temp_fun_apply = self.fun_apply(fun, value)
        return temp_fun_apply.eval()
    
    def eval(self, term : Term) -> Term:
        if isinstance(term, Var):
            return term.val
        elif isinstance(term, Apply):
            return term.eval()
        else:
            return term








'''
 Copyright 2022 Yingte Xu
 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at

 http://www.apache.org/licenses/LICENSE-2.0

 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.
'''

# ------------------------------------------------------------
# pos_info.py
#
# the class to preserve position information
# ------------------------------------------------------------

from __future__ import annotations
from typing import Any, List

class PosInfo:

    cur_file : str = ""

    def __init__(self, data : int | PosInfo | None = None, ):
        '''
        lineno = None means there is no position information
        '''
        # check for copy construction
        if isinstance(data, PosInfo):
            self.lineno = data.lineno
            self.file = data.file
        elif isinstance(data, int) or data is None:
            self.lineno : int | None = data
            self.file : str = PosInfo.cur_file
        else:
            raise Exception()

    @staticmethod
    def str(pos : PosInfo | None) -> str:
        if pos is not None:
            return str(pos)
        else:
            return ""

    def __str__(self) -> str:
        if self.lineno is None:
            return ""
        else:
            return " (" + self.file + ", line " + str(self.lineno) + ") "


'''
 Copyright 2022 Yingte Xu
 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at

 http://www.apache.org/licenses/LICENSE-2.0

 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.
'''

# ------------------------------------------------------------
# proof_tactis.py
#
# define the tacits for proof automatic generation
# ------------------------------------------------------------

from __future__ import annotations
from typing import Any, List, Dict, Tuple

from nqpv import dts
from nqpv.vsystem.log_system import RuntimeErrorWithLog

from .qvarls_term import QvarlsTerm, type_qvarls, val_qvarls
from .opt_pair_term import OptPairTerm, type_opt_pair, val_opt_pair
from . import qpre_term
from .qpre_term import QPreTerm, type_qpre, val_qpre
from .scope_term import ScopeTerm


from .proof_hint_term import AbortHintTerm, IfHintTerm, InitHintTerm, NondetHintTerm, ProofHintTerm, ProofSeqHintTerm, QPreHintTerm, SkipHintTerm, SubproofHintTerm, UnionHintTerm, UnitaryHintTerm, WhileHintTerm, val_proof_hint
from .proof_term import AbortProofTerm, IfProofTerm, InitProofTerm, NondetProofTerm, ProofSeqTerm, ProofSttTerm, ProofTerm, ProofDefinedTerm, QPreProofTerm, SkipProofTerm, SubproofTerm, UnionProofTerm, UnitaryProofTerm, WhileProofTerm


def construct_proof(hint : ProofHintTerm, pre : dts.Term, post : dts.Term, arg_ls : dts.Term, scope : ScopeTerm) -> ProofTerm:
    '''
    construct a proof term from this hint
    '''
    if not isinstance(pre, dts.Term) or not isinstance(post, dts.Term)\
            or not isinstance(arg_ls, dts.Term) or not isinstance(scope, ScopeTerm):
        raise ValueError()
    pre_val = val_qpre(pre)
    post_val = val_qpre(post)

    # need to check whether arg_ls covers the pre and the post
    arg_ls_val = val_qvarls(arg_ls)
    if not arg_ls_val.cover(pre_val.all_qvarls) or \
            not arg_ls_val.cover(post_val.all_qvarls):
        raise RuntimeErrorWithLog("The argument list '" + str(arg_ls) + "' must cover that of the precondition and the postcondition.")

    # calculate the proof statements
    proof_stts = wp_calculus(hint, post_val, scope)

    try:
        QPreTerm.sqsubseteq(pre_val, proof_stts.pre_val, scope)
    except RuntimeErrorWithLog:
        raise RuntimeErrorWithLog("The precondition of this proof does not hold.")

    return ProofDefinedTerm(pre, hint, proof_stts, post, arg_ls)

def wp_calculus(hint : ProofHintTerm, post : QPreTerm, scope : ScopeTerm) -> ProofSttTerm:
    '''
    calculate the weakest precondition, of a program given by proof_hint with respect to the post condition
    return a proof statement for this
    '''
    scope.report("wp calculus : " + str(hint.label))
    scope.report("{ ? }")
    scope.report(str(hint))
    scope.report(str(post) + "\n")
    
    if isinstance(hint, SkipHintTerm):
        return SkipProofTerm(post, post)
        
    elif isinstance(hint, AbortHintTerm):
        post_val = val_qpre(post)
        pre = qpre_term.qpre_I(post_val.all_qvarls, scope)
        return AbortProofTerm(pre, post)

    elif isinstance(hint, InitHintTerm):
        post_val = val_qpre(post)
        pre = qpre_term.qpre_init(post_val, hint.qvarls_val, scope)
        return InitProofTerm(pre, post, hint._qvarls)

    elif isinstance(hint, UnitaryHintTerm):
        post_val = val_qpre(post)        
        pre = qpre_term.qpre_contract(post_val, hint.opt_pair_val.dagger(), scope)
        return UnitaryProofTerm(pre, post, hint._opt_pair)
    
    elif isinstance(hint, IfHintTerm):
        post_val = val_qpre(post)

        if len(post_val.opt_pairs) == 1:            

            P1 = wp_calculus(hint.P1_val, post, scope)
            P0 = wp_calculus(hint.P0_val, post, scope)

            pre = qpre_term.qpre_mea_proj_sum(P0.pre_val, P1.pre_val, hint.opt_pair_val, scope)
            return IfProofTerm(pre, post, hint._opt_pair, P1, P0)

        else:
            # break the set using (Union) rule
            proof_ls : List[ProofSttTerm]= []
            union_pre = QPreTerm(())
            for pair in post_val.opt_pairs:
                pair_val = val_opt_pair(pair)
                this_post = QPreTerm((pair_val,))

                P1 = wp_calculus(hint.P1_val, this_post, scope)
                P0 = wp_calculus(hint.P0_val, this_post, scope)
                this_pre = qpre_term.qpre_mea_proj_sum(P0.pre_val, P1.pre_val, hint.opt_pair_val, scope)
                union_pre = union_pre.union(this_pre)
                proof_ls.append(IfProofTerm(this_pre, this_post, hint._opt_pair, P1, P0))
            
            return UnionProofTerm(union_pre, post, tuple(proof_ls))
        
    elif isinstance(hint, WhileHintTerm):
        post_val = val_qpre(post)

        if len(post_val.opt_pairs) == 1:            
            proposed_pre = qpre_term.qpre_mea_proj_sum(post_val, hint.inv_val, hint.opt_pair_val, scope)
            P = wp_calculus(hint.P_val, proposed_pre, scope)
            try:
                QPreTerm.sqsubseteq(hint.inv_val, P.pre_val, scope)
            except:
                raise RuntimeErrorWithLog("The predicate '" + str(hint._inv) + "' is not a valid loop invariant.")  

            return WhileProofTerm(proposed_pre, post, hint._inv, hint._opt_pair, P)
        else:
            # break the set using (Union) rule
            proof_ls : List[ProofSttTerm]= []
            union_pre = QPreTerm(())
            for pair in post_val.opt_pairs:
                pair_val = val_opt_pair(pair)
                this_post = QPreTerm((pair_val,))

                proposed_pre = qpre_term.qpre_mea_proj_sum(this_post, hint.inv_val, hint.opt_pair_val, scope)
                P = wp_calculus(hint.P_val, proposed_pre, scope)
                try:
                    QPreTerm.sqsubseteq(hint.inv_val, P.pre_val, scope)
                except:
                    raise RuntimeErrorWithLog("The predicate '" + str(hint._inv) + "' is not a valid loop invariant.")  

                union_pre = union_pre.union(proposed_pre)
                proof_ls.append(WhileProofTerm(proposed_pre, this_post, hint._inv, hint._opt_pair, P))
            
            return UnionProofTerm(union_pre, post, tuple(proof_ls))

    elif isinstance(hint, NondetHintTerm):
        proof_stts = []
        pre = QPreTerm(())
        for item in hint._proof_hints:
            item_val = val_proof_hint(item)
            new_proof_stt = wp_calculus(item_val, post, scope)
            proof_stts.append(new_proof_stt)
            pre = pre.union(new_proof_stt.pre_val)
        
        return NondetProofTerm(pre, post, tuple(proof_stts))
    
    elif isinstance(hint, SubproofHintTerm):
        cor = hint.subproof_val.arg_ls_val.get_sub_correspond(hint.arg_ls_val)
        pre_sub = hint.subproof_val.pre_val.qvar_subsitute(cor)
        post_sub = hint.subproof_val.post_val.qvar_subsitute(cor)

        try:
            QPreTerm.sqsubseteq(post_sub, post, scope)
        except RuntimeErrorWithLog:
            raise RuntimeErrorWithLog("The subproof statement '" + str(hint._subproof) + "' cannot be put here.")
        
        return SubproofTerm(pre_sub, post, hint._subproof, hint._arg_ls)
    
    elif isinstance(hint, QPreHintTerm):
        try:
            QPreTerm.sqsubseteq(hint._qpre, post, scope)
        except RuntimeErrorWithLog:
            raise RuntimeErrorWithLog("The condition hint '" + str(post) + "' does not hold.")
        
        return QPreProofTerm(hint._qpre, post, hint._qpre)

    elif isinstance(hint, UnionHintTerm):
        proof_stts = []
        pre_cal = QPreTerm(())
        post_cal = QPreTerm(())
        for item in hint._proof_hints:
            try:
                item_val = val_proof_hint(item)
                # different tactics for subproofs and proof hints
                if isinstance(item_val, ProofSeqHintTerm):
                    subhint = item_val.get_proof_hint(len(item_val._proof_hints)-1)
                    if isinstance(subhint, SubproofHintTerm):
                        item_post = subhint.subproof_val.post_val
                    elif isinstance(subhint, QPreHintTerm):
                        item_post = subhint.qpre_val
                    else:
                        raise RuntimeErrorWithLog("The postcondition of proof hint '" + str(item) + "' cannot be automatically deduced.")

                    new_proof_stt = wp_calculus(item_val, item_post, scope)
                    proof_stts.append(new_proof_stt)
                    post_cal = post_cal.union(item_post)
                    pre_cal = pre_cal.union(new_proof_stt.pre_val)

            except RuntimeErrorWithLog:
                raise RuntimeErrorWithLog("The proof '" + str(item) + "' in the union proof does not hold.")
        
        try:
            QPreTerm.sqsubseteq(post_cal, post, scope)
        except RuntimeErrorWithLog:
            raise RuntimeErrorWithLog("The postcondition and the (Union) proof hint do not fit.")

        return UnionProofTerm(pre_cal, post_cal, tuple(proof_stts))
    
    elif isinstance(hint, ProofSeqHintTerm):
        # backward transformation
        proof_stts : List[ProofSttTerm] = []
        cur_post = post
        for i in range(len(hint._proof_hints)-1, -1, -1):
            item_val = val_proof_hint(hint._proof_hints[i])
            proof_stts.insert(0, wp_calculus(item_val, cur_post, scope))
            cur_post = proof_stts[0].pre_val
        
        return ProofSeqTerm(cur_post, post, tuple(proof_stts))
    
    else:
        raise Exception()



'''
 Copyright 2022 Yingte Xu
 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at

 http://www.apache.org/licenses/LICENSE-2.0

 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.
'''

# ------------------------------------------------------------
# vkernel.py
#
# transform abstract syntax tree to semantic contents, and adjust the verification system
# ------------------------------------------------------------

from __future__ import annotations
from typing import Any, List, Dict, Tuple

import os

from nqpv import dts
from nqpv.vsystem.content.proof_tacits import construct_proof
from nqpv.vsystem.syntax.pos_info import PosInfo

from .settings import Settings
from .syntax import ast
from .syntax import vparser
from .log_system import LogSystem, RuntimeErrorWithLog
from .optenv import get_opt_env, optload, optsave

from .content.qvarls_term import QvarlsTerm
from .content.opt_pair_term import OptPairTerm
from .content.qpre_term import QPreTerm
from .content.prog_term import type_prog, AbortTerm, IfTerm, InitTerm, NondetTerm, ProgDefinedTerm, ProgDefiningTerm, ProgSttSeqTerm, ProgSttTerm, ProgTerm, SkipTerm, SubProgTerm, UnitaryTerm, WhileTerm
from .content.proof_term import ProofDefiningTerm
from .content.proof_hint_term import type_proof, AbortHintTerm, IfHintTerm, InitHintTerm, NondetHintTerm, ProofHintTerm, ProofSeqHintTerm, ProofDefinedTerm, QPreHintTerm, SkipHintTerm, SubproofHintTerm, UnionHintTerm, UnitaryHintTerm, WhileHintTerm
from .content.scope_term import ScopeTerm, VarPath


class VKernel:
    def __init__(self, name : str, folder_path : str = "", parent : VKernel | None = None):
        '''
        parent : the parent scope. if None, then the new scope will be created
        folder_path : the folder path (relative to the run path) of the scope modules. (It is usually the folder where the file is located.)
        '''
        super().__init__()
        self.cur_scope : ScopeTerm
        self.folder_path = folder_path
        if parent is None:
            self.cur_scope = ScopeTerm(name, None)
            self.cur_scope.inject(get_opt_env())
        else:
            self.cur_scope = ScopeTerm(name, parent.cur_scope)


    def inject(self, scope : ScopeTerm) -> None:
        '''
        inject the scope to the current scope
        variables with the same name will not be reassigned
        '''
        self.cur_scope.inject(scope)


    def report(self, msg : str) -> None:
        self.cur_scope.report(msg)

    @staticmethod
    def process_module(path : str) -> ScopeTerm:
        '''
        Note : there is the requirement for filename extension ".nqpv"
        '''
        folder_path = os.path.dirname(path)
        file_name = os.path.basename(path)
        index_dot = file_name.find(".")
        if file_name[index_dot:] != ".nqpv":
            raise RuntimeErrorWithLog("The file '" + path + "' is not a '.npqv' file.")

        try:
            p_prog = open(path, 'r')
            prog_str = p_prog.read()
            p_prog.close()
        except:
            raise RuntimeErrorWithLog("The file '" + path + "' not found.")
        

        module_name = file_name[:index_dot]
        kernel = VKernel("global", folder_path)

        try:
            PosInfo.cur_file = module_name
            ast_scope = vparser.parser.parse(prog_str)
            scope = kernel.eval_scope(ast_scope, module_name)
            return scope
        except RuntimeErrorWithLog:
            raise RuntimeErrorWithLog("in module '" + module_name + "'.")


    def eval_varpath(self, varpath_ast : ast.AstVar) -> VarPath:
        list = [item.id for item in varpath_ast.data]
        return VarPath(tuple(list))

    def eval_scope(self, scope_ast : ast.AstScope, name = "__local_scope") -> ScopeTerm:
        '''
        evaluate the scope using a subkernel. this allows the nested setting strategy
        '''
        new_kernel = VKernel(name, self.folder_path, self)
        # register the scope
        self.cur_scope[name] = new_kernel.cur_scope
        # enter the new setting 
        parent_setting = ScopeTerm.cur_setting 
        ScopeTerm.cur_setting = new_kernel.cur_scope.settings

        for cmd in scope_ast.cmd_ls:

            try:
                new_kernel.report("Evaluating the '" + cmd.label + "' command." + PosInfo.str(cmd.pos) + "\n")

                if isinstance(cmd, ast.AstDefinition):
                    new_kernel.eval_def(cmd)

                elif isinstance(cmd, ast.AstExample):
                    new_kernel.eval_expr(cmd.expr)

                elif isinstance(cmd, ast.AstAxiom):
                    raise NotImplementedError()

                elif isinstance(cmd, ast.AstShow):
                    # append the information
                    LogSystem.channels["info"].append("\n" + str(cmd.pos) + "\n" + str(new_kernel.eval_expr(cmd.expr)))

                elif isinstance(cmd, ast.AstSetting):
                    if cmd.setting_item == "EPS":
                        if cmd.data <= 0:
                            raise RuntimeErrorWithLog("The setting 'EPS' must be greater than 0.", cmd.pos)
                        new_kernel.cur_scope.settings.EPS = cmd.data
                    elif cmd.setting_item == "SDP_PRECISION":
                        if cmd.data <= 0:
                            raise RuntimeErrorWithLog("The setting 'SDP_PRECISION' must be greater than 0.", cmd.pos)
                        new_kernel.cur_scope.settings.SDP_precision = cmd.data
                    elif cmd.setting_item == "SILENT":
                        new_kernel.cur_scope.settings.silent = cmd.data
                    else:
                        raise Exception()

                elif isinstance(cmd, ast.AstSaveOpt):
                    var_path = new_kernel.eval_varpath(cmd.var)
                    real_path = os.path.join(self.folder_path, cmd.path)
                    optsave(new_kernel.cur_scope[var_path], real_path)
                else:
                    raise Exception()
            except RuntimeErrorWithLog:
                LogSystem.channels["error"].append("Invalid '" + cmd.label + "' command." + PosInfo.str(cmd.pos))

        # return to the parent setting
        ScopeTerm.cur_setting = parent_setting
            
        return new_kernel.cur_scope
    
    def eval_qvarls(self, qvarls : ast.AstQvarLs) -> QvarlsTerm:
        try:
            id_ls = []
            for item in qvarls.data:
                if VarPath((item.id,)) in self.cur_scope:
                    raise RuntimeErrorWithLog("The variable '" + str(item) + "' already exists, and cannot be used as a quantum variable identifier.")
                id_ls.append(item.id)

            return QvarlsTerm(tuple(id_ls))
        except RuntimeErrorWithLog:
            raise RuntimeErrorWithLog("Invalid quantum variable list.", qvarls.pos)

    def eval_qpre(self, qpre : ast.AstPredicate | ast.AstInv) -> QPreTerm:
        try:
            pair_ls : List[OptPairTerm] = []
            for pair in qpre.data:
                var_path = self.eval_varpath(pair[0])
                opt = self.cur_scope[var_path]
                qvarls = self.eval_qvarls(pair[1])
                pair_ls.append(OptPairTerm(opt, qvarls))
            return QPreTerm(tuple(pair_ls))
        except RuntimeErrorWithLog:
            raise RuntimeErrorWithLog("Invalid quantum predicate.", qpre.pos)



    def eval_prog(self, data : ast.Ast) -> ProgSttTerm:
        if isinstance(data, ast.AstSkip):
            return SkipTerm()
        elif isinstance(data, ast.AstAbort):
            return AbortTerm()
        elif isinstance(data, ast.AstInit):
            qvarls = self.eval_qvarls(data.qvar_ls)
            return InitTerm(qvarls)
        elif isinstance(data, ast.AstUnitary):
            opt = self.cur_scope[self.eval_varpath(data.opt)]
            qvarls = self.eval_qvarls(data.qvar_ls)
            pair = OptPairTerm(opt, qvarls)
            return UnitaryTerm(pair)
        elif isinstance(data, ast.AstIf):
            S1 = self.eval_prog(data.prog1)
            S0 = self.eval_prog(data.prog0)
            opt = self.cur_scope[self.eval_varpath(data.opt)]
            qvarls = self.eval_qvarls(data.qvar_ls)
            pair = OptPairTerm(opt, qvarls)
            return IfTerm(pair, S1, S0)
        elif isinstance(data, ast.AstWhile):
            S = self.eval_prog(data.prog)
            opt = self.cur_scope[self.eval_varpath(data.opt)]
            qvarls = self.eval_qvarls(data.qvar_ls)
            pair = OptPairTerm(opt, qvarls)
            return WhileTerm(pair, S)
        elif isinstance(data, ast.AstNondet):
            prog_ls = []
            for subprog in data.data:
                prog_ls.append(self.eval_prog(subprog))
            return NondetTerm(tuple(prog_ls))
        elif isinstance(data, ast.AstSubprog):
            subprog = self.cur_scope[self.eval_varpath(data.subprog)]
            qvarls = self.eval_qvarls(data.qvar_ls)
            return SubProgTerm(subprog, qvarls)
        elif isinstance(data, ast.AstProgSeq):
            prog_ls = []
            for subprog in data.data:
                prog_ls.append(self.eval_prog(subprog))
            return ProgSttSeqTerm(tuple(prog_ls))
        else:
            raise Exception()

    def eval_proof_hint(self, data : ast.Ast) -> ProofHintTerm:
        if isinstance(data, ast.AstSkip):
            return SkipHintTerm()
        elif isinstance(data, ast.AstAbort):
            return AbortHintTerm()
        elif isinstance(data, ast.AstInit):
            qvarls = self.eval_qvarls(data.qvar_ls)
            return InitHintTerm(qvarls)
        elif isinstance(data, ast.AstUnitary):
            opt = self.cur_scope[self.eval_varpath(data.opt)]
            qvarls = self.eval_qvarls(data.qvar_ls)
            pair = OptPairTerm(opt, qvarls)
            return UnitaryHintTerm(pair)
        elif isinstance(data, ast.AstIfProof):
            P1 = self.eval_proof_hint(data.proof1)
            P0 = self.eval_proof_hint(data.proof0)
            opt = self.cur_scope[self.eval_varpath(data.opt)]
            qvarls = self.eval_qvarls(data.qvar_ls)
            pair = OptPairTerm(opt, qvarls)
            return IfHintTerm(pair, P1, P0)
        elif isinstance(data, ast.AstWhileProof):
            inv = self.eval_qpre(data.inv)
            P = self.eval_proof_hint(data.proof)
            opt = self.cur_scope[self.eval_varpath(data.opt)]
            qvarls = self.eval_qvarls(data.qvar_ls)
            pair = OptPairTerm(opt, qvarls)
            return WhileHintTerm(inv, pair, P)
        elif isinstance(data, ast.AstNondetProof):
            proof_ls = []
            for subproof in data.data:
                proof_ls.append(self.eval_proof_hint(subproof))
            return NondetHintTerm(tuple(proof_ls))
        elif isinstance(data, ast.AstSubproof):
            subproof = self.cur_scope[self.eval_varpath(data.subproof)]
            qvarls = self.eval_qvarls(data.qvar_ls)
            return SubproofHintTerm(subproof, qvarls)
        elif isinstance(data, ast.AstUnionProof):
            proof_ls = []
            for subproof in data.data:
                proof_ls.append(self.eval_proof_hint(subproof))
            return UnionHintTerm(tuple(proof_ls))
        elif isinstance(data, ast.AstProofSeq):
            proof_ls = []
            for subproof in data.data:
                proof_ls.append(self.eval_proof_hint(subproof))
            return ProofSeqHintTerm(tuple(proof_ls))
        elif isinstance(data, ast.AstPredicate):
            qpre = self.eval_qpre(data)
            return QPreHintTerm(qpre)
        else:
            raise Exception()
    
    def eval_expr(self, expr : ast.AstExpression) -> dts.Term:
        '''
        this method will always return the value, not the variable
        '''
        try:
            if isinstance(expr, ast.AstExpressionVar):
                var_path = self.eval_varpath(expr.var)
                return self.cur_scope[var_path].val

            elif isinstance(expr, ast.AstExpressionValue):
                if isinstance(expr.data, ast.AstProgExpr):    
                    # definition
                    if isinstance(expr.data.data, ast.AstProgSeq):
                        if not isinstance(expr.data.type, ast.AstTypeProg):
                            raise Exception()
                        prog_seq = self.eval_prog(expr.data.data)
                        arg_ls = self.eval_qvarls(expr.data.type.qvarls)
                        return ProgDefinedTerm(prog_seq, arg_ls)
                    else:
                        raise RuntimeErrorWithLog("The expression is not of type '" + str(expr.data.type) + "'.", expr.data.pos)
                    
                elif isinstance(expr.data, ast.AstProofExpr):
                    #definition
                    if isinstance(expr.data.data, ast.AstProof):
                        if not isinstance(expr.data.type, ast.AstTypeProof):
                            raise Exception()
                        pre = self.eval_qpre(expr.data.data.pre)
                        proof_hint_seq = self.eval_proof_hint(expr.data.data.seq)
                        post = self.eval_qpre(expr.data.data.post)
                        arg_ls = self.eval_qvarls(expr.data.type.qvarls)
                        return construct_proof(proof_hint_seq, pre, post, arg_ls, self.cur_scope)
                    else:
                        raise RuntimeErrorWithLog("The expression is not of type '" + str(expr.data.type) + "'.", expr.data.pos)

                elif isinstance(expr.data, ast.AstLoadOpt):
                    real_path = os.path.join(self.folder_path, expr.data.path)
                    return optload(real_path)

                elif isinstance(expr.data, ast.AstScope):
                    return self.eval_scope(expr.data)
                else:
                    raise Exception()
            else:
                raise Exception()

        except RuntimeErrorWithLog:
            raise RuntimeErrorWithLog("Invalid expression.", expr.pos)
        
    def eval_def(self, cmd : ast.AstDefinition) -> None:
        '''
        evaluate the expression and return the value (containing type and data)
        '''
        if isinstance(cmd.expr, ast.AstExpressionVar):
            var_path = self.eval_varpath(cmd.expr.var)
            self.cur_scope[cmd.var.id] = self.cur_scope[var_path]
        elif isinstance(cmd.expr, ast.AstExpressionValue):
            if isinstance(cmd.expr.data, ast.AstProgExpr):
                try:
                    if not isinstance(cmd.expr.data.type, ast.AstTypeProg):
                        raise Exception()
                    # create the var being defined
                    self.cur_scope[cmd.var.id] = ProgDefiningTerm(self.eval_qvarls(cmd.expr.data.type.qvarls))

                    # evaluate
                    value = self.eval_expr(cmd.expr)
                    
                    # assign
                    self.cur_scope.remove_var(cmd.var.id)
                    self.cur_scope[cmd.var.id] = value

                except RuntimeErrorWithLog:
                    self.cur_scope.remove_var(cmd.var.id)
                    raise RuntimeErrorWithLog("Invalid program definition.", cmd.pos)


            elif isinstance(cmd.expr.data, ast.AstProofExpr):
                try:
                    if not isinstance(cmd.expr.data.type, ast.AstTypeProof):
                        raise Exception()
                    # create the var being defined
                    self.cur_scope[cmd.var.id] = ProofDefiningTerm(self.eval_qvarls(cmd.expr.data.type.qvarls))
                    
                    # evaluate
                    value = self.eval_expr(cmd.expr)

                    # assign
                    self.cur_scope.remove_var(cmd.var.id)
                    self.cur_scope[cmd.var.id] = value
                
                except RuntimeErrorWithLog:
                    self.cur_scope.remove_var(cmd.var.id)
                    raise RuntimeErrorWithLog("Invalid proof definition.", cmd.pos)

            elif isinstance(cmd.expr.data, ast.AstScope):                    
                subscope = self.eval_scope(cmd.expr.data, cmd.var.id)
                self.cur_scope[cmd.var.id] = subscope

            # normal assignment
            elif isinstance(cmd.expr.data, ast.AstLoadOpt):
                self.cur_scope[cmd.var.id] = self.eval_expr(cmd.expr)
            elif isinstance(cmd.expr.data, ast.AstImport):
                # get a new kernel
                module_scope = VKernel.process_module(os.path.join(self.folder_path, cmd.expr.data.path))
                self.cur_scope[cmd.var.id] = module_scope
                
            else:
                raise Exception()
        else:
            raise Exception()




'''
 Copyright 2022 Yingte Xu
 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at

 http://www.apache.org/licenses/LICENSE-2.0

 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.
'''

# ------------------------------------------------------------
# ast.py
#
# the abstract syntax tree
# ------------------------------------------------------------

from __future__ import annotations
from typing import Any, List, Tuple

from .pos_info import PosInfo

class Ast:
    def __init__(self, pos : PosInfo, ast_label : str):
        pass
        self.pos : PosInfo = pos
        self.label : str = ast_label

class AstScope(Ast):
    def __init__(self, pos : PosInfo, cmd_ls : List[Ast]):
        super().__init__(pos, "scope")
        self.cmd_ls : List[Ast] = cmd_ls

class AstDefinition(Ast):
    def __init__(self, pos : PosInfo, var : AstID, expr : AstExpression):
        super().__init__(pos, "definition")
        self.var : AstID = var
        self.expr : AstExpression = expr

class AstExample(Ast):
    def __init__(self, pos : PosInfo, expr : AstExpression):
        super().__init__(pos, "example")
        self.expr : AstExpression = expr

class AstAxiom(Ast):
    def __init__(self, pos : PosInfo, subproof : AstID, pre : AstPredicate,
        subprog : AstID, qvar_ls : AstQvarLs, post : AstPredicate):
        super().__init__(pos, "axiom")
        raise NotImplementedError()

class AstSetting(Ast):
    def __init__(self, pos : PosInfo, setting_item : str, data : Any):
        super().__init__(pos, "setting")
        self.setting_item : str = setting_item
        self.data : Any = data

class AstShow(Ast):
    def __init__(self, pos : PosInfo, expr : AstExpression):
        super().__init__(pos, "show")
        self.expr : AstExpression = expr

class AstSaveOpt(Ast):
    def __init__(self, pos : PosInfo, var : AstVar, path : str):
        super().__init__(pos, "save operator")
        self.var : AstVar = var
        self.path : str = path


class AstType(Ast):
    def __init__(self, pos : PosInfo):
        super().__init__(pos, "type")

class AstTypeScope(AstType):
    def __init__(self, pos : PosInfo):
        super().__init__(pos)

class AstTypeProg(AstType):
    def __init__(self, pos : PosInfo, qvarls : AstQvarLs):
        super().__init__(pos)
        self.qvarls : AstQvarLs = qvarls
    def __str__(self) -> str:
        return "program " + str(self.qvarls)

class AstTypeProof(AstType):
    def __init__(self, pos : PosInfo, qvarls : AstQvarLs):
        super().__init__(pos)
        self.qvarls : AstQvarLs = qvarls
    def __str__(self) -> str:
        return "proof " + str(self.qvarls)

class AstTypeOperator(AstType):
    def __init__(self, pos : PosInfo):
        super().__init__(pos)

class AstImport(Ast):
    def __init__(self, pos : PosInfo, path : str):
        super().__init__(pos, "import module")
        self.path : str = path

class AstLoadOpt(Ast):
    def __init__(self, pos : PosInfo, path : str):
        super().__init__(pos, "load operator")
        self.path : str = path

class AstExpression(Ast):
    def __init__(self, pos : PosInfo):
        super().__init__(pos, "expresssion")

class AstProofExpr(Ast):
    def __init__(self, pos : PosInfo, type : AstType, data : AstProofSeq):
        super().__init__(pos, "proof expression")
        self.type : AstType = type
        self.data : AstProofSeq = data

class AstProgExpr(Ast):
    def __init__(self, pos : PosInfo, type : AstType, data : AstProgSeq):
        super().__init__(pos, "program expression")
        self.type : AstType = type
        self.data : AstProgSeq = data

class AstExpressionValue(AstExpression):
    def __init__(self, pos : PosInfo, data : Ast):
        super().__init__(pos)
        self.data : Ast = data

class AstExpressionVar(AstExpression):
    def __init__(self, pos : PosInfo, var : AstVar):
        super().__init__(pos)
        self.var : AstVar = var

class AstQvarLs(Ast):
    def __init__(self, pos : PosInfo, data : List[AstID]):
        super().__init__(pos, "qvar list")
        self.data : List[AstID] = data
    
    def __len__(self) -> int:
        return len(self.data)

    def __str__(self) -> str:
        if len(self.data) == 0:
            return "[]"

        r = "[" + str(self.data[0])
        for i in range(1, len(self.data)):
            r += " " + str(self.data[i])
        r += "]"
        return r

class AstVar(Ast):
    def __init__(self, pos : PosInfo, data : List[AstID]):
        super().__init__(pos, "variable")
        self.data : List[AstID] = data
    
    def __str__(self) -> str:
        r = str(self.data[0])
        for i in range(1, len(self.data)):
            r += "." + str(self.data[i])
        return r

class AstID(Ast):
    def __init__(self, pos : PosInfo, id : str):
        super().__init__(pos, "ID")
        self.id : str = id
    
    def __str__(self) -> str:
        return self.id

class AstIfProof(Ast):
    def __init__(self, pos : PosInfo, opt : AstVar, qvar_ls : AstQvarLs, proof1 : AstProof, proof0 : AstProof):
        super().__init__(pos, "if proof")
        self.opt : AstVar = opt
        self.qvar_ls : AstQvarLs = qvar_ls
        self.proof1 : AstProof = proof1
        self.proof0 : AstProof = proof0

class AstInv(Ast):
    def __init__(self, pos : PosInfo, data : List[Tuple[AstVar, AstQvarLs]]):
        super().__init__(pos, "loop invariant")
        self.data : List[Tuple[AstVar, AstQvarLs]] =  data

class AstUnionProof(Ast):
    def __init__(self, pos : PosInfo, data : List[Ast]):
        super().__init__(pos, "union proof")
        self.data : List[Ast] = data

class AstSubproof(Ast):
    def __init__(self, pos : PosInfo, subproof : AstVar, qvar_ls : AstQvarLs):
        super().__init__(pos, "subproof")
        self.subproof : AstVar = subproof
        self.qvar_ls : AstQvarLs = qvar_ls
    
    def __str__(self) -> str:
        return str(self.subproof) + str(self.qvar_ls)


class AstWhileProof(Ast):
    def __init__(self, pos : PosInfo, inv : AstInv, opt : AstVar, qvar_ls : AstQvarLs, proof : AstProof):
        super().__init__(pos, "while proof")
        self.inv : AstInv = inv
        self.opt : AstVar = opt
        self.qvar_ls : AstQvarLs = qvar_ls
        self.proof : AstProof = proof

class AstNondetProof(Ast):
    def __init__(self, pos : PosInfo, data : List[AstProof]):
        super().__init__(pos, "nondeterministic choice proof")
        self.data : List[AstProof] = data

class AstPredicate(Ast):
    def __init__(self, pos : PosInfo, data : List[Tuple[AstVar, AstQvarLs]]):
        super().__init__(pos, "predicate")
        self.data : List[Tuple[AstVar, AstQvarLs]] = data


class AstProof(Ast):
    def __init__(self, pos : PosInfo, pre : AstPredicate, seq : AstProofSeq, post : AstPredicate):
        super().__init__(pos, "proof")
        self.seq : AstProofSeq = seq
        self.pre : AstPredicate = pre
        self.post : AstPredicate = post

class AstProofSeq(Ast):
    def __init__(self, pos : PosInfo, data : List[Ast]):
        super().__init__(pos, "proof sequence")
        self.data : List[Ast] = data

class AstProgSeq(Ast):
    def __init__(self, pos : PosInfo, data : List[Ast]):
        super().__init__(pos, "program sequence")
        self.data : List[Ast] = data
    

class AstNondet(Ast):
    def __init__(self, pos : PosInfo, data : List[AstProgSeq]):
        super().__init__(pos, "nondeterministic choice")
        self.data : List[AstProgSeq] = data

class AstWhile(Ast):
    def __init__(self, pos : PosInfo, opt : AstVar, qvar_ls : AstQvarLs, prog : AstProgSeq):
        super().__init__(pos, "while")
        self.opt : AstVar = opt
        self.qvar_ls : AstQvarLs = qvar_ls
        self.prog : AstProgSeq = prog

class AstIf(Ast):
    def __init__(self, pos : PosInfo, opt : AstVar, qvar_ls : AstQvarLs, prog1 : AstProgSeq, prog0 : AstProgSeq):
        super().__init__(pos, "if")
        self.opt : AstVar = opt
        self.qvar_ls : AstQvarLs = qvar_ls
        self.prog1 : AstProgSeq = prog1
        self.prog0 : AstProgSeq = prog0

class AstUnitary(Ast):
    def __init__(self, pos : PosInfo, opt : AstVar, qvar_ls : AstQvarLs):
        super().__init__(pos, "unitary transformation")
        self.opt : AstVar = opt
        self.qvar_ls : AstQvarLs = qvar_ls


class AstInit(Ast):
    def __init__(self, pos : PosInfo, qvar_ls : AstQvarLs):
        super().__init__(pos, "initialization")
        self.qvar_ls : AstQvarLs = qvar_ls
class AstAbort(Ast):
    def __init__(self, pos : PosInfo):
        super().__init__(pos, "abort")


class AstSkip(Ast):
    def __init__(self, pos : PosInfo):
        super().__init__(pos, "skip")


class AstSubprog(Ast):
    def __init__(self, pos : PosInfo, subprog : AstVar, qvar_ls : AstQvarLs):
        super().__init__(pos, "subprogram")
        self.subprog : AstVar = subprog
        self.qvar_ls : AstQvarLs = qvar_ls



'''
 Copyright 2022 Yingte Xu
 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at

 http://www.apache.org/licenses/LICENSE-2.0

 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.
'''

# ------------------------------------------------------------
# opt_pair_term.py
#
# defining the operator-variable pair terms
# ------------------------------------------------------------
from __future__ import annotations
from typing import Any, List, Tuple, Dict

from nqpv import dts
from nqpv.vsystem.log_system import RuntimeErrorWithLog

from . import opt_kernel
from .qvarls_term import QvarlsTerm, type_qvarls, val_qvarls
from .opt_term import OperatorTerm, type_operator, val_opt
from .scope_term import ScopeTerm

fac = dts.TermFact()

type_opt_pair = fac.axiom("opt_pair", fac.sort_term(0))

class OptPairTerm(dts.Term):
    def __init__(self, opt : dts.Term, qvarls : dts.Term):

        if not isinstance(opt, dts.Term) or not isinstance(qvarls, dts.Term):
            raise ValueError()

        # check the terms
        if opt.type != type_operator:
            raise RuntimeErrorWithLog("The term '" + str(opt) + "' is not an operator.")
        if qvarls.type != type_qvarls:
            raise RuntimeErrorWithLog("The term '" + str(qvarls) + "' is not a quantum variable list")
        
        opt_val : OperatorTerm = val_opt(opt)
        qvarls_val : QvarlsTerm = val_qvarls(qvarls)

        try:
            opt_val.qnum
        except ValueError:
            raise RuntimeErrorWithLog("The operator '" + str(opt) + "' can not be used here. All indices must be 2 valued.")


        # check the qubit number
        if opt_val.qnum != qvarls_val.qnum:
            raise RuntimeErrorWithLog("The operator '" + str(opt) + "' and the quantum variable list '" + 
                str(qvarls) + "' does not match on qubit numbers.")
        
        super().__init__(type_opt_pair, None)
        self._opt : dts.Term = opt
        self._qvarls : dts.Term = qvarls
    
    @property
    def opt(self) -> dts.Term:
        return self._opt
    
    @property
    def opt_val(self) -> OperatorTerm:
        return val_opt(self._opt)
    
    @property
    def qvarls(self) -> dts.Term:
        return self._qvarls

    @property
    def qvarls_val(self) -> QvarlsTerm:
        return val_qvarls(self._qvarls)

    def eval(self) -> dts.Term:
        return OptPairTerm(self.opt.eval(), self.qvarls.eval())

    @property
    def unitary_pair(self) -> bool:
        return self._opt.eval().unitary # type: ignore
    
    @property
    def hermitian_predicate_pair(self) -> bool:
        return self._opt.eval().hermitian_predicate # type: ignore
    
    @property
    def measurement_pair(self) -> bool:
        return self._opt.eval().measurement # type: ignore

    def __eq__(self, other) -> bool:
        if isinstance(other, OptPairTerm):
            return self.opt == other.opt and self.qvarls == other.qvarls
        elif isinstance(other, dts.Var):
            raise NotImplemented
        else:
            return False
    
    def __str__(self) -> str:
        return str(self.opt) + str(self.qvarls)

    def dagger(self) -> OptPairTerm:
        '''
        return the dagger operator
        '''
        return OptPairTerm(self.opt_val.dagger(), self._qvarls)

    def opt_mea0(self) -> OptPairTerm:
        '''
        return the operator for measurement result 0
        '''
        return OptPairTerm(self.opt_val.opt_mea0(), self._qvarls)
    
    def opt_mea1(self) -> OptPairTerm:
        '''
        return the operator for measurement result 1
        '''
        return OptPairTerm(self.opt_val.opt_mea1(), self._qvarls)
    
    def __add__(self, other : OptPairTerm) -> OptPairTerm:
        '''
        <automatic extension> for hermitian pairs
        '''
        if not isinstance(other, OptPairTerm):
            raise ValueError()
        if self.qvarls != other.qvarls:
            # automatic extension for hermitian pairs
            if self.hermitian_predicate_pair and other.hermitian_predicate_pair:
                all_qvarls = self.qvarls_val.join(other.qvarls_val)
                new_self = hermitian_extend(self, all_qvarls)
                new_other = hermitian_extend(other, all_qvarls)
                return OptPairTerm(new_self.opt_val + new_other.opt_val, all_qvarls)
            else:
                raise ValueError()
        else:
            return OptPairTerm(self.opt_val + other.opt_val, self.qvarls)

    def qvar_substitute(self, correspondence : Dict[str, str]) -> OptPairTerm:
        return OptPairTerm(self._opt, self.qvarls_val.qvar_substitute(correspondence))

def val_opt_pair(term : dts.Term) -> OptPairTerm:
    if not isinstance(term, dts.Term):
        raise ValueError()
    if term.type != type_opt_pair:
        raise ValueError()

    if isinstance(term, OptPairTerm):
        return term
    elif isinstance(term, dts.Var):
        val = term.val
        if not isinstance(val, OptPairTerm):
            raise Exception()
        return val
    else:
        raise Exception()

def hermitian_I(all_qvarls : QvarlsTerm) -> OptPairTerm:
    if not isinstance(all_qvarls, QvarlsTerm):
        raise ValueError()
    # produce the I operator
    m = opt_kernel.eye_tensor(len(all_qvarls))
    opt = OperatorTerm(m)
    opt.ensure_hermitian_predicate()
    opt.ensure_unitary()
    return OptPairTerm(opt, all_qvarls)


def hermitian_contract(H : OptPairTerm, M : OptPairTerm) -> OptPairTerm:
    '''
    scope : the scope to preserve the newly created operators
    <automatic extension>
    '''
    if not isinstance(H, OptPairTerm) or not isinstance(M, OptPairTerm):
        raise ValueError()
    if not H.hermitian_predicate_pair:
        raise RuntimeErrorWithLog("The operator variable pair '" + str(H) + "' is not a hermitian predicate pair.")
    
    # automatic extension
    if not H.qvarls_val.cover(M.qvarls_val):
        extended_qvarls = H.qvarls_val.join(M.qvarls_val)
        H = hermitian_extend(H, extended_qvarls)

    new_m = opt_kernel.hermitian_contract(H.qvarls_val.vls, H.opt_val.m, M.qvarls_val.vls, M.opt_val.m)
    new_opt = OperatorTerm(new_m)
    new_opt.ensure_hermitian_predicate()

    return OptPairTerm(new_opt, H.qvarls_val)

def hermitian_init(H : OptPairTerm, qvarls : QvarlsTerm) -> OptPairTerm:
    '''
    scope : the scope to preserve the newly created operators
    <automatic extension>
    '''
    if not isinstance(H, OptPairTerm) or not isinstance(qvarls, QvarlsTerm):
        raise ValueError()
    if not H.hermitian_predicate_pair:
        raise RuntimeErrorWithLog("The operator variable pair '" + str(H) + "' is not a hermitian predicate pair.")
    
    # automatic extension
    if not H.qvarls_val.cover(qvarls):
        extended_qvarls = H.qvarls_val.join(qvarls)
        H = hermitian_extend(H, extended_qvarls)

    new_m = opt_kernel.hermitian_init(H.qvarls_val.vls, H.opt_val.m, qvarls.vls)
    new_opt = OperatorTerm(new_m)
    new_opt.ensure_hermitian_predicate()

    return OptPairTerm(new_opt, H.qvarls_val)

def hermitian_extend(H : OptPairTerm, all_qvarls : QvarlsTerm) -> OptPairTerm:
    '''
    scope : the scope to preserve the newly created operators
    '''
    if not isinstance(H, OptPairTerm) or not isinstance(all_qvarls, QvarlsTerm):
        raise ValueError()
    if not H.hermitian_predicate_pair:
        raise RuntimeErrorWithLog("The operator variable pair '" + str(H) + "' is not a hermitian predicate pair.")
    
    if not all_qvarls.cover(H.qvarls_val):
        raise RuntimeErrorWithLog("The variable list of '" + str(H) + "' must be covered by the quantum variable list '" + str(all_qvarls) + "' to expand to.")

    # check whether extend is unnecessary
    if all_qvarls == H.qvarls_val:
        return H

    new_m = opt_kernel.hermitian_extend(all_qvarls.vls,  H.opt_val.m, H.qvarls_val.vls)
    new_opt = OperatorTerm(new_m)
    new_opt.ensure_hermitian_predicate()

    return OptPairTerm(new_opt, all_qvarls)


'''
 Copyright 2022 Yingte Xu
 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at

 http://www.apache.org/licenses/LICENSE-2.0

 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.
'''

# ------------------------------------------------------------
# opt_kernel.py
#
# quantum linear algebra tools needed in this verifier
# ------------------------------------------------------------
from __future__ import annotations
from typing import Any, List, Tuple

import numpy as np


from nqpv.vsystem.log_system import LogSystem
from .scope_term import ScopeTerm    

def np_eps_equal(a : np.ndarray, b : np.ndarray) -> bool:
    '''
    check whether two tensors a and b are equal, according to maximum norm.
    '''
    if a.shape != b.shape :
        return False

    diff = np.max(np_complex_norm(a - b))  # type: ignore
    return diff < ScopeTerm.cur_setting.EPS

def np_complex_norm(m : np.ndarray) -> np.ndarray:
    '''
    calculate the element wise norm
    '''
    return np.sqrt(m.real * m.real + m.imag * m.imag)


def check_unity(m : np.ndarray) -> bool:
    '''
    check whether tensor m is unitary
    m: tensor of shape (2,2,...,2), with the row indices in front of the column indices
    '''

    if len(m.shape) % 2 == 1:
        #LogSystem.channels["error"].append("The dimension is invalid for an unitary operator.")
        return False

    for dim in m.shape:
        if dim != 2:
            #LogSystem.channels["error"].append("The dimension is invalid for an unitary operator.")
            return False
    
    # calculate the dim for matrix
    dim_m : int = 2**(len(m.shape)//2)
    matrix = m.reshape((dim_m, dim_m))

    # check the equality of U^dagger @ U and I

    if not np_eps_equal(matrix @ np.transpose(np.conj(matrix)), np.eye(dim_m)):
        #LogSystem.channels["error"].append("The operator is not unitary.")
        return False
    return True


def check_hermitian_predicate(m : np.ndarray) -> bool:
    '''
    check whether tensor m is hermitian and 0 <= m <= I
    m: tensor of shape (2,2,...,2), with the row indices in front of the column indices
    '''

    if len(m.shape) % 2 == 1:
        #LogSystem.channels["error"].append("The dimension is invalid for an Hermitian predicate.")
        return False

    for dim in m.shape:
        if dim != 2:
            #LogSystem.channels["error"].append("The dimension is invalid for an Hermitian predicate.")
            return False
    
    # calculate the dim for matrix
    dim_m = 2**(len(m.shape)//2)
    matrix = m.reshape((dim_m, dim_m))

    # check the equivalence of U^dagger @ U and I
    if not np_eps_equal(matrix, np.transpose(np.conj(matrix))):
        #LogSystem.channels["error"].append("The operator is not a Hermitian operator.")
        return False

    # check 0 <= matrix <= I
    e_vals = np.linalg.eigvals(matrix)
    if np.any(e_vals < 0 - ScopeTerm.cur_setting.EPS) or np.any(e_vals > 1 + ScopeTerm.cur_setting.EPS):
        #LogSystem.channels["error"].append("The requirement 0 <= Predicate <= I is not satisfied.")
        return False
        
    return True


def check_measure(m : np.ndarray) -> bool:
    '''
    check whether tensor m is a valid measurement
    m: tensor of shape (2,2,...,2), with the row indices in front of the column indices. 
        The first index of m corresponds to measurement result 0 or 1.
    '''
        
    if len(m.shape) % 2 == 0:
        #LogSystem.channels["error"].append("The dimension is invalid for a measurement set.")
        return False

    for dim in m.shape:
        if dim != 2:
            #LogSystem.channels["error"].append("The dimension is invalid for a measurement set.")
            return False
    
    # calculate the dim for matrix
    dim_m = 2**((len(m.shape)-1)//2)

    # pick out M0 and M1
    m0 = m[0].reshape((dim_m, dim_m))
    m1 = m[1].reshape((dim_m, dim_m))

    # check the equivalence of U^dagger @ U and I
    if not np_eps_equal(m0.conj().transpose() @ m0 + m1.conj().transpose() @ m1, np.eye(dim_m)):
        #LogSystem.channels["error"].append("This tensor does not satisfy the normalization requirement of a measurement set.")
        return False
    return True

def eye_tensor(qubitn : int) -> np.ndarray:
    '''
    return the identity matrix of 'qubitn' qubits, in the form of a (2,2,2,...) tensor, row indices in the front.
    '''
    return np.eye(2**qubitn).reshape((2,)*qubitn*2)

def dagger(M : np.ndarray) -> np.ndarray:
    '''
    for a tensor M with shape (2,2,2,...), return M^dagger
    Note: M should have been already checked
    '''
    nM = len(M.shape)//2
    trans = list(range(nM, nM*2)) + list(range(0, nM))
    return np.conjugate(M).transpose(trans)

################################### operation with qvars
def hermitian_contract(qvar: Tuple[str,...], H : np.ndarray, qvar_act : Tuple[str,...], M : np.ndarray) -> np.ndarray:
    '''
    conduct the transformation M.H.M^dagger and return the result hermitian operator
    qvar: name sequence of qubits in H
    qvar_act: name sequence of qubits in M

    Note: the operators H and M should have been checked already

    [index sequence of tensor H (and M)]

            qvar == [q1, q2, q3, ... , qn]

              n  n+1 n+2 n+3     2n-2 2n-1
              |   |   |   |  ...  |   |
             ---------------------------
            | q1  q2  q3     ...      qn|
             ---------------------------
              |   |   |   |  ...  |   |
              0   1   2   3      n-2 n-1

    '''
    nH = len(qvar)
    nM = len(qvar_act)
    # decide the indices for contraction. note that M^dagger is accessed by its conjugate and the same index list iM_ls
    iM_ls = list(range(nM, 2*nM))
    iH_left_ls = [qvar.index(v) for v in qvar_act]
    iH_right_ls = [i + nH for i in iH_left_ls]

    # decide the rearrangements, since the standard rearrangement is not what we want
    count_rem_MH = 0
    count_rem_HMd = nH
    rearrange_MH = []
    rearrange_HMd = []
    for i in range(nH):
        if i in iH_left_ls:
            rearrange_MH.append(2*nH-nM + qvar_act.index(qvar[i]))
        else:
            rearrange_MH.append(count_rem_MH)
            count_rem_MH += 1
        if i + nH in iH_right_ls:
            rearrange_HMd.append(2*nH-nM + qvar_act.index(qvar[i - nH]))
        else:
            rearrange_HMd.append(count_rem_HMd)
            count_rem_HMd += 1

    rearrange_MH = rearrange_MH + list(range(nH - nM, 2*nH - nM))
    rearrange_HMd = list(range(nH)) + rearrange_HMd

    # conduct the contraction and rearrange the indices
    temp1 = np.tensordot(H, M, (iH_left_ls, iM_ls)).transpose(rearrange_MH)
    temp2 = np.tensordot(temp1, np.conjugate(M), (iH_right_ls, iM_ls)).transpose(rearrange_HMd)
    return temp2    

def hermitian_init(qvar: Tuple[str,...], H : np.ndarray, qvar_init: Tuple[str,...]) -> np.ndarray:
    '''
    initialize hermitian operator H at variables 'qvar_init'
    '''
    P0 = np.array([[1., 0.],
                    [0., 0.]])
    P1 = np.array([[0., 0.],
                    [1., 0.]])
    # initialize all the variables in order
    tempH = H
    for var in qvar_init:
        a = hermitian_contract(qvar, tempH, (var,), P0)
        b = hermitian_contract(qvar, tempH, (var,), P1)
        tempH = a + b
    
    return tempH

def tensor_to_matrix(t : np.ndarray) -> np.ndarray:
    nM = len(t.shape)//2
    ndim = 2**nM
    return t.reshape((ndim, ndim))


def hermitian_extend(qvar: Tuple[str,...], H : np.ndarray, qvar_H: Tuple[str,...]) -> np.ndarray:
    '''
    extend the given hermitian operator, according to all variables qvar, and return
    '''
    nAll = len(qvar)
    nH = len(qvar_H)
    m_I = eye_tensor(nAll - nH)

    temp = np.tensordot(H, m_I, ([],[]))

    # rearrange the indices
    count_ext = 0
    r_left = []
    r_right = []
    for i in range(nAll):
        if qvar[i] in qvar_H:
            pos = qvar_H.index(qvar[i])
            r_left.append(pos)
            r_right.append(nH + pos)
        else:
            r_left.append(2*nH + count_ext)
            r_right.append(nAll + nH + count_ext)
            count_ext += 1
    
    return temp.transpose(r_left + r_right)


def get_opt_qnum(m : np.ndarray) -> int:
    if not isinstance(m, np.ndarray):
        raise Exception()
    
    for dim in m.shape:
        if dim != 2:
            raise ValueError()

    if len(m.shape) % 2 == 1:
        return (len(m.shape) - 1)//2
    else:
        return len(m.shape)//2



'''
 Copyright 2022 Yingte Xu
 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at

 http://www.apache.org/licenses/LICENSE-2.0

 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.
'''

# ------------------------------------------------------------
# scope_term.py
#
# defines the (inductive) naming scope terms and their methods
# ------------------------------------------------------------

from __future__ import annotations
from typing import Any, List, Dict, Tuple

from nqpv import dts
from nqpv.vsystem.log_system import RuntimeErrorWithLog
from nqpv.vsystem.settings import Settings

fac = dts.TermFact()

type_scope = fac.axiom("scope", fac.sort_term(0))

class VarPath:
    def __init__(self, path : Tuple[str,...], pointer : int = 0):
        if not isinstance(path, tuple):
            raise ValueError()
        for item in path:
            if not isinstance(item, str):
                raise ValueError()
        if len(path) == 0:
            raise ValueError()

        self.path : Tuple[str,...] = path
        self._pointer = pointer
    
    @property
    def current(self) -> str:
        return self.path[self._pointer]

    def postfix(self) -> VarPath | None:
        '''
        return the postfix VarPath (pointer starting from the next string)
        '''
        if self._pointer < len(self.path) - 1:
            return VarPath(self.path, self._pointer + 1)
        else:
            return None
    
    def __str__(self) -> str:
        r = str(self.path[0])
        for i in range(1, len(self.path)):
            r += "." + str(self.path[i])
        return r


class ScopeTerm(dts.Term):


    cur_setting : Settings = Settings()

    def __init__(self, label : str, parent_scope : dts.Term | None):
        if not (isinstance(parent_scope, dts.Term) or parent_scope is None) or not isinstance(label, str):
            raise ValueError()
        

        super().__init__(type_scope, None)

        # auto naming number for every scope
        self.auto_naming_no : int = 0

        if parent_scope is not None:
            if parent_scope.type != type_scope:
                raise ValueError()
            parent_scope = parent_scope.eval()
            if not isinstance(parent_scope, ScopeTerm):
                raise Exception()
        self._parent_scope : ScopeTerm | None = parent_scope

        # the label of this environment
        self._label : str = label
        # dictionary to store all the variables in this environment
        # note: the dictionary is the local id, and the dict value (var) has a polished id
        self._vars : Dict[str, dts.Var] = {}
        self.settings = Settings()

    @property
    def parent_scope(self) -> ScopeTerm | None:
        return self._parent_scope

    @property
    def scope_prefix(self) -> str:
        if self._parent_scope is None:
            return self._label + "."
        else:
            return self._parent_scope.scope_prefix + self._label + "."

    def __str__(self) -> str:
        r = "<scope " + self.scope_prefix + ">\n"
        r += str(self.settings) + "\n"
        for key in self._vars:
            r += "\t" + key + "\t\t" + str(self._vars[key].type) + "\n"
        return r

    def _search(self, label : str) -> dts.Var | None:
        if label in self._vars:
            return self._vars[label]
        elif self._parent_scope is not None:
            return self._parent_scope._search(label)
        # may found global itself
        elif self._label == label:
            return dts.Var(label, type_scope, self)
        else:
            return None


    def __getitem__(self, key : VarPath | str) -> dts.Var:
        if isinstance(key, str):
            return self[VarPath((key,))]

        find_res = self._search(key.current)
        if find_res is not None:
            next_key = key.postfix()
            if next_key is None:
                # if the search is over
                return find_res
            else:
                if find_res.type == type_scope:
                    # search in the next scope
                    next_scope_val = find_res.val
                    if not isinstance(next_scope_val, ScopeTerm):
                        raise Exception()
                    return next_scope_val[next_key]
                
        raise RuntimeErrorWithLog("The variable '" + str(key) + "' is not defined.")


    def __setitem__(self, key : str, value : dts.Term) -> None:
        '''
        Note: value will be packaged into a Var term, with the polished name
        also used to assign new variables
        '''
        if not isinstance(value, dts.Term):
            raise ValueError()
        
        self._vars[key] = dts.Var(self.scope_prefix + key, value.type, value, key)
    
    def _search_value(self, value : dts.Term, id_used : set[str]) -> str | None :
        '''
        id_used : preserved the searched ids in child scopes to avoid the references to overlapped old variables.
        '''
        for key in self._vars:
            if key in id_used:
                continue
            id_used.add(key)
            if self[key] == value:
                return key
        
        if self.parent_scope is None:
            return None
        else:
            return self.parent_scope._search_value(value, id_used)

    def append(self, value : dts.Term) -> str:
        '''
        check whether the value already exists in this environment.
        If yes, return the variable.
        If not, create a new variable with an auto name and return the name used.
        '''
        search_res = self._search_value(value, set())
        if search_res is None:
            name = self.auto_name()
            self[name] = value
            return name
        else:
            return search_res


    def remove_var(self, key : str) -> None:
        '''
        find the variable in this scope and remove it
        '''
        if key not in self._vars:
            raise RuntimeErrorWithLog("The variable '" + key + "' dost not exist in this scope, and can not be deleted.")
        self._vars.pop(key)

    def __contains__(self, key : VarPath | str) -> bool:
        if isinstance(key, str):
            return VarPath((key,)) in self

        find_res = self._search(key.current)
        if find_res is not None:
            next_key = key.postfix()
            if next_key is None:
                # if the search is over
                return True
            else:
                if find_res.type == type_scope:
                    # search in the next scope
                    next_scope_val = find_res.val
                    if not isinstance(next_scope_val, ScopeTerm):
                        raise Exception()
                    return next_key in next_scope_val 
        return False

    def inject(self, var_env : ScopeTerm) -> None:
        '''
        inject the var environment to the current var environment
        (variables with the same name will be reassigned)
        '''
        for var in var_env._vars:
            self[var] = var_env[var].val
    
    def auto_name(self, naming_prefix = "VAR") -> str:
        '''
        return an auto name, which does not appear in this environment
        '''
        r = naming_prefix + str(self.auto_naming_no)
        self.auto_naming_no += 1
        # ensure that the new name will not overlap any current variable
        while r in self._vars:
            r = naming_prefix + str(self.auto_naming_no)
            self.auto_naming_no += 1
        return r

    def report(self, msg : str) -> None:
        if not self.settings.silent:
            print(msg)


'''
 Copyright 2022 Yingte Xu
 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at

 http://www.apache.org/licenses/LICENSE-2.0

 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.
'''

# ------------------------------------------------------------
# qvarls_term.py
#
# defining the quantum variable list terms
# ------------------------------------------------------------
from __future__ import annotations
from typing import Any, List, Tuple, Dict

from nqpv import dts

from nqpv.vsystem.log_system import RuntimeErrorWithLog

fac = dts.TermFact()

type_qvarls = fac.axiom("qvarls", fac.sort_term(0))

class QvarlsTerm(dts.Term):
    def __init__(self, qvarls : Tuple[str,...]):
        #check the terms
        if not isinstance(qvarls, tuple):
            raise ValueError()

        #repeat check
        appeared = set()
        for item in qvarls:
            if not isinstance(item, str):
                raise ValueError()

            if item in appeared:
                raise RuntimeErrorWithLog("The variable '" + item + "' repeats in the list '" + str(qvarls) + "'.")
            else:
                appeared.add(item)

        super().__init__(type_qvarls, None)
        self._qvarls : Tuple[str,...] = qvarls
    
    @property
    def vls(self) -> Tuple[str,...]:
        return self._qvarls

    @property
    def qnum(self) -> int:
        return len(self._qvarls)
    
    def __len__(self) -> int:
        return len(self._qvarls)
    
    def __eq__(self, other) -> bool:
        if isinstance(other, QvarlsTerm):
            return self._qvarls == other._qvarls
        elif isinstance(other, dts.Var):
            raise NotImplemented
        else:
            return False

    def __str__(self) -> str:
        if len(self) == 0:
            return "[]"
        r = "[" + self._qvarls[0]
        for i in range(1, len(self)):
            r += " " + self._qvarls[i]
        r += "]"
        return r


    def get_sub_correspond(self, arg_ls : QvarlsTerm) -> Dict[str, str]:
        '''
        get the substitution correspondence (arg_ls as arguments) for qvar_substitution
        '''
        if not isinstance(arg_ls, QvarlsTerm):
            raise ValueError()
        if self.qnum != arg_ls.qnum:
            raise ValueError()

        cor : Dict[str, str] = {}
        for i in range(self.qnum):
            cor[self._qvarls[i]] = arg_ls._qvarls[i]
        return cor
        
    
    def qvar_substitute(self, correspondence : Dict[str, str]) -> QvarlsTerm:
        if not isinstance(correspondence, dict):
            raise ValueError()
        new_qvarls = []
        for qvar in self._qvarls:
            if qvar not in correspondence:
                raise ValueError()
            new_qvar = correspondence[qvar]
            if not isinstance(new_qvar, str):
                raise ValueError()
            new_qvarls.append(new_qvar)
        
        return QvarlsTerm(tuple(new_qvarls))

    def cover(self, other : QvarlsTerm) -> bool:
        '''
        return whether this qvar list 'covers' the other qvar list
        '''
        for qvar in other._qvarls:
            if qvar not in self._qvarls:
                return False
        return True

    
    def join(self, other : QvarlsTerm) -> QvarlsTerm:
        '''
        return the new qvarls term, which joins the new variables in 'other'
        at the end of 'self' list
        '''
        new_qvarls = list(self._qvarls)
        for qvar in other._qvarls:
            if qvar not in new_qvarls:
                new_qvarls.append(qvar)
        return QvarlsTerm(tuple(new_qvarls))

def val_qvarls(term : dts.Term) -> QvarlsTerm:
    if not isinstance(term, dts.Term):
        raise ValueError()
    if term.type != type_qvarls:
        raise ValueError()
        
    if isinstance(term, QvarlsTerm):
        return term
    elif isinstance(term, dts.Var):
        val = term.val
        if not isinstance(val, QvarlsTerm):
            raise Exception()
        return val
    else:
        raise Exception()




'''
 Copyright 2022 Yingte Xu
 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at

 http://www.apache.org/licenses/LICENSE-2.0

 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.
'''

# ------------------------------------------------------------
# vparser.py
#
# parser
# ------------------------------------------------------------

from __future__ import annotations
from typing import Any, List

import ply.yacc as yacc

from nqpv.vsystem.log_system import LogSystem, RuntimeErrorWithLog

from .pos_info import PosInfo
from .vlexer import tokens, lexer
from . import ast


# program declaration section
def p_scope(p):
    '''
    scope   : cmd
            | scope cmd
    '''
    if isinstance(p[1], ast.AstScope):
        p[0] = ast.AstScope(p[1].pos, p[1].cmd_ls + [p[2]])
    else:
        p[0] = ast.AstScope(p[1].pos, [p[1]])

    if p[0] is None:
        raise Exception()

def p_cmd(p):
    '''
    cmd     : definition
            | example
            | axiom
            | show
            | setting
            | save
    '''
    p[0] = p[1]

    if p[0] is None:
        raise Exception()

def p_save(p):
    '''
    save    : SAVE var AT STRING END
    '''
    p[0] = ast.AstSaveOpt(PosInfo(p.slice[1].lineno), p[2], p[4][1:-1])

    if p[0] is None:
        raise Exception()

def p_setting(p):
    '''
    setting : SETTING EPS ASSIGN FLOAT_NUM END
            | SETTING SDP_PRECISION ASSIGN FLOAT_NUM END
            | SETTING SILENT ASSIGN TRUE END
            | SETTING SILENT ASSIGN FALSE END
    '''

    # FLOAT_NUM is already checked
    if p[2] == "EPS":
        p[0] = ast.AstSetting(PosInfo(p.slice[1].lineno), p[2], float(p[4]))
    elif p[2] == "SDP_PRECISION":
        p[0] = ast.AstSetting(PosInfo(p.slice[1].lineno), p[2], float(p[4]))
    elif p[2] == "SILENT":
        if p[4] == "true":
            p[0] = ast.AstSetting(PosInfo(p.slice[1].lineno), p[2], True)
        else:
            p[0] = ast.AstSetting(PosInfo(p.slice[1].lineno), p[2], False)
    else:
        raise Exception()

    if p[0] is None:
        raise Exception()


def p_show(p):
    '''
    show    : SHOW expression END
    '''
    p[0] = ast.AstShow(PosInfo(p.slice[1].lineno), p[2])

    if p[0] is None:
        raise Exception()

def p_definition(p):
    '''
    definition  : DEF id ASSIGN expression END
    '''
    p[0] = ast.AstDefinition(PosInfo(p.slice[1].lineno), p[2], p[4])

    if p[0] is None:
        raise Exception()

def p_example(p):
    '''
    example : EXAMPLE ASSIGN expression END
    '''
    p[0] = ast.AstExample(PosInfo(p.slice[1].lineno), p[3])

    if p[0] is None:
        raise Exception()

def p_axiom(p):
    '''
    axiom   : AXIOM id ':' predicate PROGRAM qvar_ls predicate END
    '''
    p[0] = ast.AstAxiom(PosInfo(p.slice[1].lineno), p[2], p[4], p[5], p[6], p[7])

    if p[0] is None:
        raise Exception()

def p_type(p):
    '''
    type    : PROGRAM qvar_ls
            | PROOF qvar_ls
            | SCOPE
            | OPERATOR
    '''

    if p[1] == "program":
        p[0] = ast.AstTypeProg(PosInfo(p.slice[1].lineno), p[2])
    elif p[1] == "proof":
        p[0] = ast.AstTypeProof(PosInfo(p.slice[1].lineno), p[2])
    elif p[1] == "scope":
        p[0] = ast.AstTypeScope(PosInfo(p.slice[1].lineno))
    elif p[1] == "operator":
        p[0] = ast.AstTypeOperator(PosInfo(p.slice[1].lineno))

    if p[0] is None:
        raise Exception()

def p_expression(p):
    '''
    expression  : expr_data
                | var
    '''
    if isinstance(p[1], ast.AstVar):
        p[0] = ast.AstExpressionVar(p[1].pos, p[1])
    else:
        p[0] = ast.AstExpressionValue(p[1].pos, p[1])

    if p[0] is None:
        raise Exception()

def p_expr_data(p):
    '''
    expr_data   : type ':' prog
                | type ':' proof
                | scope
                | load
                | import
    '''
    if len(p) > 2:
        if isinstance(p[3], ast.AstProgSeq):
            p[0] = ast.AstProgExpr(p[1].pos, p[1], p[3])
        elif isinstance(p[3], ast.AstProof):
            p[0] = ast.AstProofExpr(p[1].pos, p[1], p[3])
    else:
        p[0] = p[1]

    if p[0] is None:
        raise Exception()

def p_import(p):
    '''
    import  : IMPORT STRING
    '''
    p[0] = ast.AstImport(PosInfo(p.slice[1].lineno), p[2][1:-1])

    if p[0] is None:
        raise Exception()


def p_load(p):
    '''
    load    : LOAD STRING
    '''
    p[0] = ast.AstLoadOpt(PosInfo(p.slice[1].lineno), p[2][1:-1])

    if p[0] is None:
        raise Exception()

def p_qvar_ls(p):
    '''
    qvar_ls : qvar_ls_pre ']'
    '''
    p[0] = p[1]

    if p[0] is None:
        raise Exception()

def p_qvar_ls_pre(p):
    '''
    qvar_ls_pre : '[' id
                | qvar_ls_pre id
    '''
    if p[1] == '[':
        p[0] = ast.AstQvarLs(PosInfo(p.slice[1].lineno), [p[2]])
    else:
        p[0] = ast.AstQvarLs(p[1].pos, p[1].data + [p[2]])

    if p[0] is None:
        raise Exception()

def p_predicate(p):
    '''
    predicate   : predicate_pre '}'
    '''
    p[0] = p[1]

    if p[0] is None:
        raise Exception()

def p_predicate_pre(p):
    '''
    predicate_pre   : '{' var qvar_ls
                    | predicate_pre var qvar_ls
    '''
    if p[1] == '{':
        p[0] = ast.AstPredicate(PosInfo(p.slice[1].lineno), [(p[2], p[3])])
    else:
        p[0] = ast.AstPredicate(p[1].pos, p[1].data + [(p[2], p[3])])

    if p[0] is None:
        raise Exception()


def p_prog(p):
    '''
    prog    : statement
            | prog ';' statement
    '''
    if len(p) == 2:
        p[0] = ast.AstProgSeq(p[1].pos, [p[1]])
    else:
        p[0] = ast.AstProgSeq(p[1].pos, p[1].data + [p[3]])

    if p[0] is None:
        raise Exception()

def p_statement(p):
    '''
    statement : skip
                | abort
                | init
                | unitary
                | if
                | while
                | nondet
                | var qvar_ls
    '''
    if len(p) == 2:
        p[0] = p[1]
    else:
        p[0] = ast.AstSubprog(p[1].pos, p[1], p[2])

    if p[0] is None:
        raise Exception()

def p_skip(p):
    '''
    skip    : SKIP
    '''
    p[0] = ast.AstSkip(PosInfo(p.slice[1].lineno))

    if p[0] is None:
        raise Exception()

def p_abort(p):
    '''
    abort   : ABORT
    '''
    p[0] = ast.AstAbort(PosInfo(p.slice[1].lineno))

    if p[0] is None:
        raise Exception()

def p_init(p):
    '''
    init    : id INIT
            | qvar_ls INIT
    '''
    if isinstance(p[1], ast.AstQvarLs):
        p[0] = ast.AstInit(p[1].pos, p[1])
    else:
        qvar_ls = ast.AstQvarLs(p[1].pos, [p[1]])
        p[0] = ast.AstInit(qvar_ls.pos, qvar_ls)

    if p[0] is None:
        raise Exception()

def p_unitary(p):
    '''
    unitary : id MUL_EQ var
            | qvar_ls MUL_EQ var
    '''
    if isinstance(p[1], ast.AstQvarLs):
        p[0] = ast.AstUnitary(p[1].pos, p[3], p[1])
    else:
        qvar_ls = ast.AstQvarLs(p[1].pos, [p[1]])
        p[0] = ast.AstUnitary(qvar_ls.pos, p[3], qvar_ls)

    if p[0] is None:
        raise Exception()

def p_if(p):
    '''
    if      : IF var qvar_ls THEN prog ELSE prog END
    '''
    p[0] = ast.AstIf(PosInfo(p.slice[1].lineno), p[2], p[3], p[5], p[7])

    if p[0] is None:
        raise Exception()

def p_while(p):
    '''
    while   : WHILE var qvar_ls DO prog END
    '''
    p[0] = ast.AstWhile(PosInfo(p.slice[1].lineno), p[2], p[3], p[5])

    if p[0] is None:
        raise Exception()

def p_nondet(p):
    '''
    nondet  : nondet_pre '#' prog ')'
    '''
    p[0] = ast.AstNondet(p[1].pos, p[1].data + [p[3]])

    if p[0] is None:
        raise Exception()

def p_nondet_pre(p):
    '''
    nondet_pre  : '(' prog
                | nondet_pre '#' prog
    '''
    if len(p) == 3:
        p[0] = ast.AstNondet(PosInfo(p.slice[1].lineno), [p[2]])
    else:
        p[0] = ast.AstNondet(p[1].pos, p[1].data + [p[3]])

    if p[0] is None:
        raise Exception()

def p_proof(p):
    '''
    proof   : predicate ';' proof_mid ';' predicate
    '''
    p[0] = ast.AstProof(p[1].pos, p[1], p[3], p[5])

    if p[0] is None:
        raise Exception()

def p_proof_mid(p):
    '''
    proof_mid   : proof_statement
                | proof_mid ';' proof_statement
    '''
    if len(p) == 2:
        p[0] = ast.AstProofSeq(p[1].pos, [p[1]])
    else:
        p[0] = ast.AstProofSeq(p[1].pos, p[1].data + [p[3]])

    if p[0] is None:
        raise Exception()

def p_proof_statement(p):
    '''
    proof_statement : var qvar_ls
                    | skip
                    | abort
                    | init
                    | unitary
                    | if_proof
                    | while_proof
                    | nondet_proof
                    | union_proof
                    | predicate
    '''
    if isinstance(p[1], ast.AstVar):
        p[0] = ast.AstSubproof(p[1].pos, p[1], p[2])
    else:
        p[0] = p[1]

    if p[0] is None:
        raise Exception()

def p_if_proof(p):
    '''
    if_proof    : IF var qvar_ls THEN proof_mid ELSE proof_mid END
    '''
    p[0] = ast.AstIfProof(PosInfo(p.slice[1].lineno), p[2], p[3], p[5], p[7])

    if p[0] is None:
        raise Exception()

def p_while_proof(p):
    '''
    while_proof : inv ';' WHILE var qvar_ls DO proof_mid END
    '''
    p[0] = ast.AstWhileProof(p[1].pos, p[1], p[4], p[5], p[7])

    if p[0] is None:
        raise Exception()

def p_nondet_proof(p):
    '''
    nondet_proof  : nondet_proof_pre '#' proof_mid ')'
    '''
    p[0] = ast.AstNondetProof(p[1].pos, p[1].data + [p[3]])

    if p[0] is None:
        raise Exception()

def p_nondet_proof_pre(p):
    '''
    nondet_proof_pre    : '(' proof_mid
                        | nondet_proof_pre '#' proof_mid
    '''
    if len(p) == 3:
        p[0] = ast.AstNondetProof(PosInfo(p.slice[1].lineno), [p[2]])
    else:
        p[0] = ast.AstNondetProof(p[1].pos, p[1].data + [p[3]])

    if p[0] is None:
        raise Exception()

def p_union_proof(p):
    '''
    union_proof : union_proof_pre ',' proof_mid ')'
    '''
    p[0] = ast.AstUnionProof(p[1].pos, p[1].data + [p[3]])

    if p[0] is None:
        raise Exception()

def p_union_proof_pre(p):
    '''
    union_proof_pre : '(' proof_mid
                    | union_proof_pre ',' proof_mid
    '''
    if len(p) == 3:
        p[0] = ast.AstUnionProof(PosInfo(p.slice[1].lineno), [p[2]])
    else:
        p[0] = ast.AstUnionProof(p[1].pos, p[1].data + [p[3]])

    if p[0] is None:
        raise Exception()


def p_inv(p):
    '''
    inv     : inv_pre '}'
    '''
    p[0] = p[1]

    if p[0] is None:
        raise Exception()

def p_inv_pre(p):
    '''
    inv_pre : '{' INV ':' var qvar_ls
            | inv_pre var qvar_ls
    '''
    if p[1] == '{':
        p[0] = ast.AstInv(PosInfo(p.slice[1].lineno), [(p[4], p[5])])
    else:
        p[0] = ast.AstInv(p[1].pos, p[1].data + [(p[2], p[3])])

    if p[0] is None:
        raise Exception()

def p_var(p):
    '''
    var     : id
            | var_pre id
    '''
    if len(p) == 2:
        p[0] = ast.AstVar(p[1].pos, [p[1]])
    else:
        p[0] = ast.AstVar(p[1].pos, p[1].data + [p[2]])

    if p[0] is None:
        raise Exception()

def p_var_pre(p):
    '''
    var_pre : id '.'
            | var_pre id '.'
    '''
    if len(p) == 3:
        p[0] = ast.AstVar(p[1].pos, [p[1]])
    else:
        p[0] = ast.AstVar(p[1].pos, p[1].data + [p[2]])

    if p[0] is None:
        raise Exception()

def p_id (p):
    '''
    id      : ID
    '''
    p[0] = ast.AstID(PosInfo(p.slice[1].lineno), p[1])
    
    if p[0] is None:
        raise Exception()

def p_error(p):
    if p is None:
        raise RuntimeErrorWithLog("unexpected end of file")
    raise RuntimeErrorWithLog("Syntax error in input: '" + str(p.value) + "'.", PosInfo(p.lineno))


# Build the lexer
parser = yacc.yacc()